"""
A module containing functions for simulating various signal impairments.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import linear
from .._farrow import FarrowResampler
from .._helper import convert_output, export, verify_arraylike, verify_only_one_specified, verify_scalar
from .._measurement import average_power


@export
def awgn(
    x: npt.ArrayLike,
    snr: float | None = None,
    noise: float | None = None,
    seed: int | None = None,
) -> npt.NDArray:
    r"""
    Adds additive white Gaussian noise (AWGN) to the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to which AWGN is added.
        snr: The desired signal-to-noise ratio (SNR) in dB. If specified, the average signal power is measured
            explicitly. It is assumed that $x[n]$ contains signal only. If the signal power is known, the
            desired noise variance can be computed and passed in `noise`. If `snr` is `None`,
            `noise` must be specified.
        noise: The noise power (variance) in linear units. If `noise` is `None`, `snr` must be specified.
        seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.

    Returns:
        The noisy signal $x[n] + w[n]$.

    Notes:
        The signal-to-noise ratio (SNR) is defined as

        $$
        \text{SNR} = \frac{P_{\text{signal,avg}}}{P_{\text{noise}}}
        = \frac{\frac{1}{N} \sum_{n=0}^{N-1} \left| x[n] \right|^2}{\sigma^2} ,
        $$

        where $\sigma^2$ is the noise variance. The output signal, with the specified SNR, is $y[n] = x[n] + w[n]$.

        For real signals:
        $$w \sim \mathcal{N}(0, \sigma^2)$$

        For complex signals:
        $$w \sim \mathcal{CN}(0, \sigma^2) = \mathcal{N}(0, \sigma^2 / 2) + j\mathcal{N}(0, \sigma^2 / 2)$$

    Examples:
        Create a real sinusoid and set its $S/N$ to 10 dB.

        .. ipython:: python

            x = np.sin(2 * np.pi * 5 * np.arange(100) / 100); \
            y = sdr.awgn(x, snr=10)

            @savefig sdr_awgn_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, label="$x[n]$"); \
            sdr.plot.time_domain(y, label="$y[n]$"); \
            plt.title("Input signal $x[n]$ and noisy output signal $y[n]$ with 10 dB SNR");

        Create a QPSK reference signal and set its $E_s/N_0$ to 10 dB. When the signal has 1 sample per symbol,
        $E_s/N_0$ is equivalent to the discrete-time $S/N$.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, psk.order, 1_000); \
            x = psk.map_symbols(s); \
            y = sdr.awgn(x, snr=10)

            @savefig sdr_awgn_2.png
            plt.figure(); \
            sdr.plot.constellation(x, label="$x[n]$", zorder=2); \
            sdr.plot.constellation(y, label="$y[n]$", zorder=1); \
            plt.title(f"QPSK constellations for $x[n]$ with $\infty$ dB $E_s/N_0$\nand $y[n]$ with 10 dB $E_s/N_0$");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True)
    verify_only_one_specified(snr, noise)

    if snr is not None:
        verify_scalar(snr, float=True)
        snr_linear = linear(snr)
        signal_power = average_power(x)
        noise_power = signal_power / snr_linear
    elif noise is not None:
        verify_scalar(noise, float=True, non_negative=True)
        noise_power = noise

    rng = np.random.default_rng(seed)
    if np.iscomplexobj(x):
        w = rng.normal(0, np.sqrt(noise_power / 2), x.shape) + 1j * rng.normal(0, np.sqrt(noise_power / 2), x.shape)
    else:
        w = rng.normal(0, np.sqrt(noise_power), x.shape)

    y = x + w

    return convert_output(y)


@export
def iq_imbalance(x: npt.ArrayLike, amplitude: float, phase: float = 0.0) -> npt.NDArray[np.complex128]:
    r"""
    Applies IQ imbalance to the complex time-domain signal $x[n]$.

    Arguments:
        x: The complex time-domain signal $x[n]$ to which IQ imbalance is applied.
        amplitude: The amplitude imbalance $A$ in dB. A positive value indicates that the in-phase component is
            larger than the quadrature component.
        phase: The phase imbalance $\phi$ in degrees. A positive value indicates that the quadrature component
            leads the in-phase component.

    Returns:
        The signal $x[n]$ with IQ imbalance applied.

    Notes:
        The IQ imbalance is applied as follows.

        $$g_I = 10^{(A/2)/20} \exp\left(j \frac{-\phi}{2} \frac{\pi}{180}\right)$$
        $$g_Q = 10^{(-A/2)/20} \exp\left(j \frac{\phi}{2} \frac{\pi}{180}\right)$$
        $$y[n] = g_I \cdot x_I[n] + j \cdot g_Q \cdot x_Q[n]$$

    Examples:
        Positive amplitude imbalance horizontally stretches the constellation, while negative amplitude imbalance
        vertically stretches the constellation.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, 4, 1_000); \
            x = psk.map_symbols(s); \
            y1 = sdr.iq_imbalance(x, 5, 0); \
            y2 = sdr.iq_imbalance(x, -5, 0)

            @savefig sdr_iq_imbalance_1.png
            plt.figure(); \
            plt.subplot(1, 2, 1); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y1, label="$y_1[n]$"); \
            plt.title("5 dB amplitude imbalance"); \
            plt.subplot(1, 2, 2); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y2, label="$y_2[n]$"); \
            plt.title("-5 dB amplitude imbalance");

        Positive phase imbalance stretches to the northwest, while negative phase imbalance stretches to the
        northeast

        .. ipython:: python

            y1 = sdr.iq_imbalance(x, 0, 20); \
            y2 = sdr.iq_imbalance(x, 0, -20)

            @savefig sdr_iq_imbalance_2.png
            plt.figure(); \
            plt.subplot(1, 2, 1); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y1, label="$y_1[n]$"); \
            plt.title("20 deg phase imbalance"); \
            plt.subplot(1, 2, 2); \
            sdr.plot.constellation(x, label="$x[n]$"); \
            sdr.plot.constellation(y2, label="$y_2[n]$"); \
            plt.title("-20 deg phase imbalance");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, imaginary=True)
    verify_scalar(amplitude, float=True)
    verify_scalar(phase, float=True)

    phase = np.deg2rad(phase)

    # TODO: Should the phase be negative for I?
    gain_i = linear(0.5 * amplitude, type="voltage") * np.exp(1j * -0.5 * phase)
    gain_q = linear(-0.5 * amplitude, type="voltage") * np.exp(1j * 0.5 * phase)

    y = gain_i * x.real + 1j * gain_q * x.imag

    return convert_output(y)


@export
def sample_rate_offset(
    x: npt.ArrayLike,
    offset: float,
    # offset: npt.ArrayLike,
    # offset_rate: npt.ArrayLike = 0.0,
    sample_rate: float = 1.0,
) -> npt.NDArray:
    r"""
    Applies a sample rate offset to the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to which the sample rate offset is applied.
        offset: The sample rate offset $\Delta f_s = f_{s,\text{new}} - f_{s,\text{old}}$ in samples/s.
        sample_rate: The sample rate $f_s$ in samples/s.

    Returns:
        The signal $x[n]$ with sample rate offset applied.

    Examples:
        Create a QPSK reference signal.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, psk.order, 1_000); \
            x = psk.map_symbols(s)

        Add 10 ppm of sample rate offset.

        .. ipython:: python

            y = sdr.sample_rate_offset(x, 10e-6)

            @savefig sdr_sample_rate_offset_1.png
            plt.figure(); \
            sdr.plot.constellation(x, label="$x[n]$", zorder=2); \
            sdr.plot.constellation(y, label="$y[n]$", zorder=1); \
            plt.title("10 ppm sample rate offset");

        Add 100 ppm of sample rate offset.

        .. ipython:: python

            y = sdr.sample_rate_offset(x, 100e-6)

            @savefig sdr_sample_rate_offset_2.png
            plt.figure(); \
            sdr.plot.constellation(x, label="$x[n]$", zorder=2); \
            sdr.plot.constellation(y, label="$y[n]$", zorder=1); \
            plt.title("100 ppm sample rate offset");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    verify_scalar(offset, float=True)
    verify_scalar(sample_rate, float=True, positive=True)

    # offset = np.asarray(offset)
    # if not (offset.ndim == 0 or offset.shape == x.shape):
    #     raise ValueError(f"Argument 'offset' must be scalar or have shape {x.shape}, not {offset.shape}.")

    # offset_rate = np.asarray(offset_rate)
    # if not (offset_rate.ndim == 0 or offset_rate.shape == x.shape):
    #     raise ValueError(f"Argument 'offset_rate' must be scalar or have shape {x.shape}, not {offset_rate.shape}.")

    rate = (sample_rate + offset) / sample_rate

    # TODO: Add ppm_rate
    # if ppm_rate:
    #     rate += ppm_rate * 1e-6 * np.arange(x.size)

    farrow = FarrowResampler()
    y = farrow(x, rate)

    return convert_output(y)


@export
def frequency_offset(
    x: npt.ArrayLike,
    offset: npt.ArrayLike,
    offset_rate: npt.ArrayLike = 0.0,
    phase: npt.ArrayLike = 0.0,
    sample_rate: float = 1.0,
) -> npt.NDArray:
    r"""
    Applies a frequency and phase offset to the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to which the frequency offset is applied.
        offset: The frequency offset $\Delta f_c = f_{c,\text{new}} - f_{c,\text{old}}$ in Hz.
        offset_rate: The frequency offset rate $\Delta f_c / \Delta t$ in Hz/s.
        phase: The phase offset $\phi$ in degrees.
        sample_rate: The sample rate $f_s$ in samples/s.

    Returns:
        The signal $x[n]$ with frequency offset applied.

    Examples:
        Create a reference signal with a constant frequency of 1 cycle per 100 samples.

        .. ipython:: python

            x = np.exp(1j * 2 * np.pi / 100 * np.arange(100))

        Add a frequency offset of 1 cycle per 100 samples (the length of the signal). Notice that the signal now
        rotates through 2 cycles instead of 1.

        .. ipython:: python

            freq = 1 / 100
            y = sdr.frequency_offset(x, freq)

            @savefig sdr_frequency_offset_1.png
            plt.figure(); \
            sdr.plot.time_domain(np.unwrap(np.angle(x)) / (2 * np.pi), label="$x[n]$"); \
            sdr.plot.time_domain(np.unwrap(np.angle(y)) / (2 * np.pi), label="$y[n]$"); \
            plt.ylabel("Absolute phase (cycles)"); \
            plt.title("Constant frequency offset (linear phase)");

        Add a frequency rate of change of 2 cycles per 100^2 samples. Notice that the signal now rotates through
        4 cycles instead of 2.

        .. ipython:: python

            freq_rate = 2 / 100**2
            y = sdr.frequency_offset(x, freq, freq_rate)

            @savefig sdr_frequency_offset_2.png
            plt.figure(); \
            sdr.plot.time_domain(np.unwrap(np.angle(x)) / (2 * np.pi), label="$x[n]$"); \
            sdr.plot.time_domain(np.unwrap(np.angle(y)) / (2 * np.pi), label="$y[n]$"); \
            plt.ylabel("Absolute phase (cycles)"); \
            plt.title("Linear frequency offset (quadratic phase)");

    Group:
        simulation-impairments
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    offset = verify_arraylike(offset, float=True)
    offset_rate = verify_arraylike(offset_rate, float=True)
    phase = verify_arraylike(phase, float=True)
    verify_scalar(sample_rate, float=True, positive=True)

    t = np.arange(x.size) / sample_rate  # Time vector in seconds
    f = offset + offset_rate * t  # Frequency vector in Hz
    lo = np.exp(1j * (2 * np.pi * f * t + np.deg2rad(phase)))  # Local oscillator
    y = x * lo  # Apply frequency offset

    return convert_output(y)



import numpy as np

import sdr


def test_impulse_response():
    iir = sdr.LeakyIntegrator(0.9)
    b_truth = [0.1]
    a_truth = [1, -0.9]
    assert np.allclose(iir.b_taps, b_truth)
    assert np.allclose(iir.a_taps, a_truth)



import numpy as np

import sdr


def test_4():
    h = sdr.half_sine(4)
    h_truth = np.sqrt(2) / 2 * np.sin(np.pi * np.arange(4) / 4)
    np.testing.assert_array_almost_equal(h, h_truth)


def test_4_3():
    h = sdr.half_sine(4, span=3)
    h_truth = np.concatenate(
        (
            np.zeros(4),
            np.sqrt(2) / 2 * np.sin(np.pi * np.arange(4) / 4),
            np.zeros(4),
        )
    )
    np.testing.assert_array_almost_equal(h, h_truth)



"""
A module containing functions to compute theoretical detection performance.
"""

from __future__ import annotations

from functools import lru_cache

import numpy as np
import numpy.typing as npt
import scipy.stats
from typing_extensions import Literal

from .._conversion import db as to_db
from .._conversion import linear
from .._helper import (
    convert_output,
    export,
    verify_arraylike,
    verify_bool,
    verify_literal,
    verify_not_specified,
    verify_scalar,
)
from .._probability import sum_distribution


@export
def h0(
    sigma2: float = 1.0,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> scipy.stats.rv_continuous:
    r"""
    Computes the statistical distribution under the null hypothesis $\mathcal{H}_0$.

    Arguments:
        sigma2: The noise variance $\sigma^2$ in linear units.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.

    Returns:
        The distribution under the null hypothesis $\mathcal{H}_0$.

    See Also:
        sdr.plot.detector_pdfs

    Examples:
        .. ipython:: python

            snr = 5  # Signal-to-noise ratio in dB
            sigma2 = 1  # Noise variance
            p_fa = 1e-1  # Probability of false alarm

        Compare the detection statistics for complex signals.

        .. ipython:: python

            A2 = sdr.linear(snr) * sigma2; A2  # Signal power, A^2

            x_h0 = rng.normal(size=100_000, scale=np.sqrt(sigma2 / 2)) + 1j * rng.normal(size=100_000, scale=np.sqrt(sigma2 / 2)); \
            x_h1 = np.sqrt(A2) + x_h0

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold
            p_d = sdr.p_d(snr, p_fa, detector); p_d

            z_h0 = np.real(x_h0); \
            z_h1 = np.real(x_h1)

            @savefig sdr_h0_1.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Coherent Detector: Probability density functions");

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold
            p_d = sdr.p_d(snr, p_fa, detector); p_d

            z_h0 = np.abs(x_h0); \
            z_h1 = np.abs(x_h1)

            @savefig sdr_h0_2.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Linear Detector: Probability density functions");

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold
            p_d = sdr.p_d(snr, p_fa, detector); p_d

            z_h0 = np.abs(x_h0) ** 2; \
            z_h1 = np.abs(x_h1) ** 2

            @savefig sdr_h0_3.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Square-Law Detector: Probability density functions");

        Compare the detection statistics for real signals.

        .. ipython:: python

            A2 = sdr.linear(snr) * sigma2; A2  # Signal power, A^2

            x_h0 = rng.normal(size=100_000, scale=np.sqrt(sigma2)); \
            x_h1 = np.sqrt(A2) + x_h0

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector, complex=False); \
            h1 = sdr.h1(snr, sigma2, detector, complex=False)

            threshold = sdr.threshold(p_fa, sigma2, detector, complex=False); threshold
            p_d = sdr.p_d(snr, p_fa, detector, complex=False); p_d

            z_h0 = np.real(x_h0); \
            z_h1 = np.real(x_h1)

            @savefig sdr_h0_4.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Coherent Detector: Probability density functions");

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector, complex=False); \
            h1 = sdr.h1(snr, sigma2, detector, complex=False)

            threshold = sdr.threshold(p_fa, sigma2, detector, complex=False); threshold
            p_d = sdr.p_d(snr, p_fa, detector, complex=False); p_d

            z_h0 = np.abs(x_h0); \
            z_h1 = np.abs(x_h1)

            @savefig sdr_h0_5.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Linear Detector: Probability density functions");

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector, complex=False); \
            h1 = sdr.h1(snr, sigma2, detector, complex=False)

            threshold = sdr.threshold(p_fa, sigma2, detector, complex=False); threshold
            p_d = sdr.p_d(snr, p_fa, detector, complex=False); p_d

            z_h0 = np.abs(x_h0) ** 2; \
            z_h1 = np.abs(x_h1) ** 2

            @savefig sdr_h0_6.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold, x=np.linspace(0, 15, 1001)); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.xlim(0, 15); \
            plt.ylim(0, 0.5); \
            plt.title("Square-Law Detector: Probability density functions");

    Group:
        detection-theory
    """
    verify_scalar(sigma2, float=True, non_negative=True)
    verify_literal(detector, ["coherent", "linear", "square-law"])
    verify_bool(complex)
    verify_scalar(n_c, int=True, positive=True)
    verify_scalar(n_nc, optional=True, int=True, positive=True)
    if detector == "coherent":
        verify_not_specified(n_nc)

    return _h0(sigma2, detector, complex, n_c, n_nc)


@lru_cache
def _h0(
    sigma2: float = 1.0,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> scipy.stats.rv_continuous:
    # sigma2 = float(sigma2)
    if n_nc is None:
        n_nc = 1

    if complex:
        nu = 2  # Degrees of freedom
        sigma2_per = sigma2 / 2  # Noise variance per dimension
    else:
        nu = 1
        sigma2_per = sigma2

    # Coherent integration scales the noise power by n_c
    sigma2_per *= n_c

    with np.errstate(invalid="ignore"):
        if detector == "coherent":
            h0 = scipy.stats.norm(0, np.sqrt(sigma2_per))
        elif detector == "linear":
            h0 = scipy.stats.chi(nu, scale=np.sqrt(sigma2_per))
            h0 = sum_distribution(h0, n_nc)
        elif detector == "square-law":
            h0 = scipy.stats.chi2(nu * n_nc, scale=sigma2_per)

    return h0


@export
def h1(
    snr: float,
    sigma2: float = 1.0,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> scipy.stats.rv_continuous:
    r"""
    Computes the statistical distribution under the alternative hypothesis $\mathcal{H}_1$.

    Arguments:
        snr: The signal-to-noise ratio $S / \sigma^2$ in dB.
        sigma2: The noise variance $\sigma^2$ in linear units.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.

    Returns:
        The distribution under the alternative hypothesis $\mathcal{H}_1$.

    See Also:
        sdr.plot.detector_pdfs

    Examples:
        .. ipython:: python

            snr = 5  # Signal-to-noise ratio in dB
            sigma2 = 1  # Noise variance
            p_fa = 1e-1  # Probability of false alarm

        Compare the detection statistics for complex signals.

        .. ipython:: python

            A2 = sdr.linear(snr) * sigma2; A2  # Signal power, A^2

            x_h0 = rng.normal(size=100_000, scale=np.sqrt(sigma2 / 2)) + 1j * rng.normal(size=100_000, scale=np.sqrt(sigma2 / 2)); \
            x_h1 = np.sqrt(A2) + x_h0

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold
            p_d = sdr.p_d(snr, p_fa, detector); p_d

            z_h0 = np.real(x_h0); \
            z_h1 = np.real(x_h1)

            @savefig sdr_h1_1.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Coherent Detector: Probability density functions");

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold
            p_d = sdr.p_d(snr, p_fa, detector); p_d

            z_h0 = np.abs(x_h0); \
            z_h1 = np.abs(x_h1)

            @savefig sdr_h1_2.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Linear Detector: Probability density functions");

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold
            p_d = sdr.p_d(snr, p_fa, detector); p_d

            z_h0 = np.abs(x_h0) ** 2; \
            z_h1 = np.abs(x_h1) ** 2

            @savefig sdr_h1_3.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Square-Law Detector: Probability density functions");

        Compare the detection statistics for real signals.

        .. ipython:: python

            A2 = sdr.linear(snr) * sigma2; A2  # Signal power, A^2

            x_h0 = rng.normal(size=100_000, scale=np.sqrt(sigma2)); \
            x_h1 = np.sqrt(A2) + x_h0

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector, complex=False); \
            h1 = sdr.h1(snr, sigma2, detector, complex=False)

            threshold = sdr.threshold(p_fa, sigma2, detector, complex=False); threshold
            p_d = sdr.p_d(snr, p_fa, detector, complex=False); p_d

            z_h0 = np.real(x_h0); \
            z_h1 = np.real(x_h1)

            @savefig sdr_h1_4.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Coherent Detector: Probability density functions");

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector, complex=False); \
            h1 = sdr.h1(snr, sigma2, detector, complex=False)

            threshold = sdr.threshold(p_fa, sigma2, detector, complex=False); threshold
            p_d = sdr.p_d(snr, p_fa, detector, complex=False); p_d

            z_h0 = np.abs(x_h0); \
            z_h1 = np.abs(x_h1)

            @savefig sdr_h1_5.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.title("Linear Detector: Probability density functions");

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector, complex=False); \
            h1 = sdr.h1(snr, sigma2, detector, complex=False)

            threshold = sdr.threshold(p_fa, sigma2, detector, complex=False); threshold
            p_d = sdr.p_d(snr, p_fa, detector, complex=False); p_d

            z_h0 = np.abs(x_h0) ** 2; \
            z_h1 = np.abs(x_h1) ** 2

            @savefig sdr_h1_6.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold, x=np.linspace(0, 15, 1001)); \
            plt.gca().set_prop_cycle(None); \
            plt.hist(z_h0, bins=101, histtype="step", density=True); \
            plt.hist(z_h1, bins=101, histtype="step", density=True); \
            plt.xlim(0, 15); \
            plt.ylim(0, 0.5); \
            plt.title("Square-Law Detector: Probability density functions");

    Group:
        detection-theory
    """
    verify_scalar(snr, float=True)
    verify_scalar(sigma2, float=True, non_negative=True)
    verify_literal(detector, ["coherent", "linear", "square-law"])
    verify_bool(complex)
    verify_scalar(n_c, int=True, positive=True)
    verify_scalar(n_nc, optional=True, int=True, positive=True)
    if detector == "coherent":
        verify_not_specified(n_nc)

    return _h1(snr, sigma2, detector, complex, n_c, n_nc)


@lru_cache
def _h1(
    snr: float,
    sigma2: float = 1.0,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> scipy.stats.rv_continuous:
    # snr = float(snr)
    # sigma2 = float(sigma2)
    if n_nc is None:
        n_nc = 1

    A2 = linear(snr) * sigma2  # Signal power, A^2
    if complex:
        nu = 2  # Degrees of freedom
        sigma2_per = sigma2 / 2  # Noise variance per dimension
    else:
        nu = 1
        sigma2_per = sigma2

    # Coherent integration scales the signal power by n_c^2 and the noise power by n_c
    A2 *= n_c**2
    sigma2_per *= n_c

    lambda_ = A2 / (sigma2_per)  # Non-centrality parameter

    with np.errstate(invalid="ignore"):
        if detector == "coherent":
            h1 = scipy.stats.norm(np.sqrt(A2), np.sqrt(sigma2_per))
        elif detector == "linear":
            if complex:
                # Rice distribution has 2 degrees of freedom
                h1 = scipy.stats.rice(np.sqrt(lambda_), scale=np.sqrt(sigma2_per))

                # Sometimes this distribution has so much SNR that SciPy throws an error when computing the mean, etc.
                # We need to check for that condition. If true, we cant approximate the distribution by a Gaussian,
                # which doesn't suffer from the same error.
                if np.isnan(h1.mean()):
                    h1 = scipy.stats.norm(np.sqrt(lambda_ * sigma2_per), scale=np.sqrt(sigma2_per))
            else:
                # Folded normal distribution has 1 degree of freedom
                h1 = scipy.stats.foldnorm(np.sqrt(lambda_), scale=np.sqrt(sigma2_per))
            h1 = sum_distribution(h1, n_nc)
        elif detector == "square-law":
            h1 = scipy.stats.ncx2(nu * n_nc, lambda_ * n_nc, scale=sigma2_per)

    return h1


@export
def p_d(
    snr: npt.ArrayLike,
    p_fa: npt.ArrayLike,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the theoretical probability of detection $P_d$.

    Arguments:
        snr: The signal-to-noise ratio $S / \sigma^2$ in dB.
        p_fa: The probability of false alarm $P_{fa}$ in $(0, 1)$.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.

    Returns:
        The probability of detection $P_d$ in $(0, 1)$.

    Notes:
        The probability of detection $P_d$ is defined as the probability that the detector output $T(x)$ exceeds the
        detection threshold $\gamma$ given that the signal is present.

        $$P_d = P\left(T(x) > \gamma \mid \mathcal{H}_1\right)$$

    Examples:
        .. ipython:: python

            snr = 3  # Signal-to-noise ratio in dB
            sigma2 = 1  # Noise variance
            p_fa = 1e-1  # Probability of false alarm

        Compute the probability of detection for the coherent detector. Plot the PDFs and observe the theoretical
        $P_d$ is achieved.

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold

            p_d = sdr.p_d(snr, p_fa, detector); p_d

            @savefig sdr_p_d_1.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, h1=h1, threshold=threshold); \
            plt.title("Coherent Detector: Probability density functions");

        Compute the probability of detection for the linear detector. Plot the PDFs and observe the theoretical
        $P_d$ is achieved.

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold

            p_d = sdr.p_d(snr, p_fa, detector); p_d

            @savefig sdr_p_d_2.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, h1=h1, threshold=threshold); \
            plt.title("Linear Detector: Probability density functions");

        Compute the probability of detection for the square-law detector. Plot the PDFs and observe the theoretical
        $P_d$ is achieved.

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold

            p_d = sdr.p_d(snr, p_fa, detector); p_d

            @savefig sdr_p_d_3.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, h1=h1, threshold=threshold); \
            plt.title("Square-Law Detector: Probability density functions");

        Compare the detection performance of each detector.

        .. ipython:: python

            p_fa = 1e-6  # Probability of false alarm

            @savefig sdr_p_d_4.png
            plt.figure(); \
            snr = np.linspace(0, 20, 101); \
            sdr.plot.p_d(snr, sdr.p_d(snr, p_fa, "coherent"), label="Coherent"); \
            sdr.plot.p_d(snr, sdr.p_d(snr, p_fa, "linear"), label="Linear"); \
            sdr.plot.p_d(snr, sdr.p_d(snr, p_fa, "square-law"), label="Square-Law"); \
            plt.legend(title="Detector"); \
            plt.title("Probability of detection");

        Compare the $P_d$ of the square-law detector for various $P_{fa}$.

        .. ipython:: python

            plt.figure(); \
            snr = np.linspace(-10, 20, 101);
            for p_fa in [1e-15, 1e-12, 1e-9, 1e-6, 1e-3, 1e-1]:
                p_d = sdr.p_d(snr, p_fa)
                sdr.plot.p_d(snr, p_d, label=f"{p_fa:1.0e}")
            @savefig sdr_p_d_5.png
            plt.legend(title="$P_{fa}$", loc="upper left"); \
            plt.title("Square-Law Detector: Probability of detection");

        Compare the receiver operating characteristics (ROCs) of the square-law detector for various SNRs.

        .. ipython:: python

            plt.figure(); \
            p_fa = np.logspace(-15, 0, 101);
            for snr in [-5, 0, 5, 10, 15, 20]:
                p_d = sdr.p_d(snr, p_fa)
                sdr.plot.roc(p_fa, p_d, label=f"{snr} dB")
            @savefig sdr_p_d_6.png
            plt.legend(title="SNR"); \
            plt.title("Square-Law Detector: Receiver operating characteristic");

    Group:
        detection-theory
    """
    snr = verify_arraylike(snr, float=True)
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    verify_literal(detector, ["coherent", "linear", "square-law"])
    verify_bool(complex)
    verify_scalar(n_c, int=True, positive=True)
    verify_scalar(n_nc, optional=True, int=True, positive=True)
    if detector == "coherent":
        verify_not_specified(n_nc)

    sigma2 = 1

    @np.vectorize
    def _calculate(snr, p_fa):
        h1 = _h1(snr, sigma2, detector, complex, n_c, n_nc)
        gamma = threshold(p_fa, sigma2, detector, complex, n_c, n_nc)
        return h1.sf(gamma)

    p_d = _calculate(snr, p_fa)

    return convert_output(p_d)


@export
def p_fa(
    threshold: npt.ArrayLike,
    sigma2: npt.ArrayLike = 1,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the theoretical probability of false alarm $P_{fa}$.

    Arguments:
        threshold: The detection threshold $\gamma$ in linear units.
        sigma2: The noise variance $\sigma^2$ in linear units.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.

    Returns:
        The probability of false alarm $P_{fa}$ in $(0, 1)$.

    Notes:
        The probability of false alarm $P_{fa}$ is defined as the probability that the detector output $T(x)$ exceeds
        the detection threshold $\gamma$ given that the signal is absent.

        $$P_{fa} = P\left(T(x) > \gamma \mid \mathcal{H}_0\right)$$

    Examples:
        .. ipython:: python

            threshold = 1.0  # Detection threshold
            sigma2 = 1  # Noise variance

        Compute the probability of false alarm for the coherent detector. Plot the PDFs and observe the theoretical
        $P_{fa}$ is achieved.

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector)

            p_fa = sdr.p_fa(threshold, sigma2, detector); p_fa

            @savefig sdr_p_fa_1.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, threshold=threshold); \
            plt.title("Coherent Detector: Probability density functions");

        Compute the probability of false alarm for the linear detector. Plot the PDFs and observe the theoretical
        $P_{fa}$ is achieved.

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector)

            p_fa = sdr.p_fa(threshold, sigma2, detector); p_fa

            @savefig sdr_p_fa_2.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, threshold=threshold); \
            plt.title("Linear Detector: Probability density functions");

        Compute the probability of false alarm for the square-law detector. Plot the PDFs and observe the theoretical
        $P_{fa}$ is achieved.

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector)

            p_fa = sdr.p_fa(threshold, sigma2, detector); p_fa

            @savefig sdr_p_fa_3.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, threshold=threshold); \
            plt.title("Square-Law Detector: Probability density functions");

    Group:
        detection-theory
    """
    threshold = verify_arraylike(threshold, float=True)
    sigma2 = verify_arraylike(sigma2, float=True, non_negative=True)
    verify_literal(detector, ["coherent", "linear", "square-law"])
    verify_bool(complex)
    verify_scalar(n_c, int=True, positive=True)
    verify_scalar(n_nc, optional=True, int=True, positive=True)
    if detector == "coherent":
        verify_not_specified(n_nc)

    @np.vectorize
    def _calculate(threshold, sigma2):
        h0 = _h0(sigma2, detector, complex, n_c, n_nc)
        return h0.sf(threshold)

    p_fa = _calculate(threshold, sigma2)

    return convert_output(p_fa)


@export
def threshold(
    p_fa: npt.ArrayLike,
    sigma2: npt.ArrayLike = 1,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the theoretical detection threshold $\gamma$.

    Arguments:
        p_fa: The desired probability of false alarm $P_{fa}$ in $(0, 1)$.
        sigma2: The noise variance $\sigma^2$ in linear units.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.

    Returns:
        The detection threshold $\gamma$ in linear units.

    Examples:
        .. ipython:: python

            p_fa = 1e-1  # Probability of false alarm
            sigma2 = 1  # Noise variance

        Compute the detection threshold for the coherent detector. Plot the PDFs and observe the desired $P_{fa}$
        is achieved.

        .. ipython:: python

            detector = "coherent"; \
            h0 = sdr.h0(sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold

            @savefig sdr_threshold_1.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, threshold=threshold); \
            plt.title("Coherent Detector: Probability density functions");

        Compute the detection threshold for the linear detector. Plot the PDFs and observe the desired $P_{fa}$
        is achieved.

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold

            @savefig sdr_threshold_2.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, threshold=threshold); \
            plt.title("Linear Detector: Probability density functions");

        Compute the detection threshold for the square-law detector. Plot the PDFs and observe the desired $P_{fa}$
        is achieved.

        .. ipython:: python

            detector = "square-law"; \
            h0 = sdr.h0(sigma2, detector)

            threshold = sdr.threshold(p_fa, sigma2, detector); threshold

            @savefig sdr_threshold_3.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0=h0, threshold=threshold); \
            plt.title("Square-Law Detector: Probability density functions");

    Group:
        detection-theory
    """
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    sigma2 = verify_arraylike(sigma2, float=True, non_negative=True)
    verify_literal(detector, ["coherent", "linear", "square-law"])
    verify_bool(complex)
    verify_scalar(n_c, int=True, positive=True)
    verify_scalar(n_nc, optional=True, int=True, positive=True)
    if detector == "coherent":
        verify_not_specified(n_nc)

    @np.vectorize
    def _calculate(p_fa, sigma2):
        h0 = _h0(sigma2, detector, complex, n_c, n_nc)
        return h0.isf(p_fa)

    threshold = _calculate(p_fa, sigma2)

    return convert_output(threshold)


@export
def threshold_factor(
    p_fa: npt.ArrayLike,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
    db: bool = False,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the theoretical detection threshold factor $\alpha$.

    Arguments:
        p_fa: The desired probability of false alarm $P_{fa}$ in $(0, 1)$.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.
        db: Indicates whether to return the detection threshold $\alpha$ in dB.

    Returns:
        The detection threshold factor $\alpha$.

    Notes:
        The detection threshold factor $\alpha$ is defined as the ratio of the detection threshold $\gamma$ to the
        mean of the detector output under the null hypothesis. This is true for linear and square-law detectors.

        $$\alpha = \frac{\gamma}{\frac{1}{N} \sum_{i=1}^{N}\{T(x[i]) \mid \mathcal{H}_0\}}$$

        For coherent detectors, the detection threshold factor $\alpha$ is defined as the ratio of the detection
        threshold $\gamma$ to the mean of the square of the detector output under the null hypothesis.
        This is required because the mean of the coherent detector output is zero.

        $$\alpha = \frac{\gamma}{\frac{1}{N} \sum_{i=1}^{N}\{T(x[i])^2 \mid \mathcal{H}_0\}}$$

    Examples:
        .. ipython:: python

            p_fa = np.logspace(-16, -1, 101)  # Probability of false alarm

            @savefig sdr_threshold_factor_1.png
            plt.figure(); \
            plt.semilogx(p_fa, sdr.threshold_factor(p_fa, detector="coherent"), label="Coherent"); \
            plt.semilogx(p_fa, sdr.threshold_factor(p_fa, detector="linear"), label="Linear"); \
            plt.semilogx(p_fa, sdr.threshold_factor(p_fa, detector="square-law"), label="Square-Law"); \
            plt.xlabel("Probability of false alarm, $P_{fa}$"); \
            plt.ylabel(r"Detection threshold factor, $\alpha$"); \
            plt.legend(title="Detector"); \
            plt.title("Detection threshold factor across false alarm rate");

    Group:
        detection-theory
    """
    sigma2 = 1
    gamma = threshold(p_fa, sigma2, detector, complex, n_c, n_nc)
    null_hypothesis = h0(sigma2, detector, complex, n_c, n_nc)

    if detector == "coherent":
        # Since mean is zero, the variance is equivalent to the mean of the square
        alpha = gamma / null_hypothesis.var()
    else:
        alpha = gamma / null_hypothesis.mean()

    if db:
        alpha = to_db(alpha)

    return alpha


@export
def min_snr(
    p_d: npt.ArrayLike,
    p_fa: npt.ArrayLike,
    detector: Literal["coherent", "linear", "square-law"] = "square-law",
    complex: bool = True,
    n_c: int = 1,
    n_nc: int | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the minimum input signal-to-noise ratio (SNR) required to achieve the desired probability of detection
    $P_d$.

    Arguments:
        p_d: The desired probability of detection $P_d$ in $(0, 1)$.
        p_fa: The probability of false alarm $P_{fa}$ in $(0, 1)$.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        n_c: The number of samples to coherently integrate $N_c$.
        n_nc: The number of samples to non-coherently integrate $N_{nc}$. Non-coherent integration is only allowable
            for linear and square-law detectors.

    Returns:
        The minimum signal-to-noise ratio (SNR) required to achieve the desired $P_d$.

    See Also:
        sdr.albersheim

    Examples:
        Compute the minimum required SNR to achieve $P_d = 0.9$ and $P_{fa} = 10^{-6}$ with a square-law detector.

        .. ipython:: python

            sdr.min_snr(0.9, 1e-6, detector="square-law")

        Now suppose the signal is non-coherently integrated $N_{nc} = 10$ times. Notice the minimum required SNR
        decreases, but by less than 10 dB. This is because non-coherent integration is less efficient than coherent
        integration.

        .. ipython:: python

            sdr.min_snr(0.9, 1e-6, detector="square-law", n_nc=10)

        Now suppose the signal is coherently integrated for $N_c = 10$ samples before the square-law detector.
        Notice the SNR now decreases by exactly 10 dB.

        .. ipython:: python

            sdr.min_snr(0.9, 1e-6, detector="square-law", n_c=10, n_nc=10)

        Compare the theoretical minimum required SNR using linear and square-law detectors.

        .. ipython:: python

            p_d = 0.9; \
            p_fa = np.logspace(-12, -1, 21)

            @savefig sdr_min_snr_1.png
            plt.figure(); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=1, detector="square-law"), label=1); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=2, detector="square-law"), label=2); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=4, detector="square-law"), label=4); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=8, detector="square-law"), label=8); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=16, detector="square-law"), label=16); \
            plt.semilogx(p_fa, sdr.min_snr(p_d, p_fa, n_nc=32, detector="square-law"), label=32); \
            plt.legend(title="$N_{nc}$"); \
            plt.xlabel("Probability of false alarm, $P_{fa}$"); \
            plt.ylabel("Minimum required input SNR (dB)"); \
            plt.title("Minimum required input SNR across non-coherent combinations for $P_d = 0.9$");

    Group:
        detection-theory
    """
    p_d = verify_arraylike(p_d, float=True, inclusive_min=0, inclusive_max=1)
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    verify_literal(detector, ["coherent", "linear", "square-law"])
    verify_bool(complex)
    verify_scalar(n_c, int=True, positive=True)
    verify_scalar(n_nc, optional=True, int=True, positive=True)
    if detector == "coherent":
        verify_not_specified(n_nc)

    calc_p_d = globals()["p_d"]

    @np.vectorize
    def _calculate(p_d, p_fa):
        def _objective(snr):
            return p_d - calc_p_d(snr, p_fa, detector, complex, n_c, n_nc)

        # The max SNR may return p_d = NaN. If so, we need to reduce the max value or optimize.brentq() will error.
        min_snr = -100  # dB
        max_snr = 30  # dB
        while True:
            if np.isnan(_objective(max_snr)):
                max_snr -= 10
            else:
                break

        snr = scipy.optimize.brentq(_objective, min_snr, max_snr)

        return snr

    snr = _calculate(p_d, p_fa)

    return convert_output(snr)



import numpy as np

import sdr


def test_4():
    h = sdr.rectangular(4)
    h_truth = 0.5 * np.ones(4)
    assert np.array_equal(h, h_truth)


def test_4_3():
    h = sdr.rectangular(4, span=3)
    h_truth = np.concatenate(
        (
            np.zeros(4),
            0.5 * np.ones(4),
            np.zeros(4),
        )
    )
    assert np.array_equal(h, h_truth)



"""
A subpackage for various measurement routines.
"""

from ._distance import *
from ._energy import *
from ._errors import *
from ._modulation import *
from ._power import *
from ._voltage import *



import numpy as np

import sdr


def test_simple():
    interleaver = sdr.BlockInterleaver(3, 4)
    assert len(interleaver) == 12

    map = np.array([0, 4, 8, 1, 5, 9, 2, 6, 10, 3, 7, 11])
    assert np.array_equal(interleaver.map, map)

    inverse_map = np.array([0, 3, 6, 9, 1, 4, 7, 10, 2, 5, 8, 11])
    assert np.array_equal(interleaver.inverse_map, inverse_map)

    x_truth = np.arange(24)
    y_truth = np.array([0, 3, 6, 9, 1, 4, 7, 10, 2, 5, 8, 11, 12, 15, 18, 21, 13, 16, 19, 22, 14, 17, 20, 23])

    y = interleaver.interleave(x_truth)
    assert np.array_equal(y, y_truth)

    x = interleaver.deinterleave(y)
    assert np.array_equal(x, x_truth)



import sdr


def test_output():
    hexdump = sdr.hexdump(b"The quick brown fox jumps over the lazy dog")
    hexdump_truth = "00000000  54 68 65 20 71 75 69 63 6b 20 62 72 6f 77 6e 20  The quick brown \n00000010  66 6f 78 20 6a 75 6d 70 73 20 6f 76 65 72 20 74  fox jumps over t\n00000020  68 65 20 6c 61 7a 79 20 64 6f 67                 he lazy dog\n"
    assert hexdump == hexdump_truth

    hexdump = sdr.hexdump([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], width=4)
    hexdump_truth = "00000000  01 02 03 04  ....\n00000004  05 06 07 08  ....\n00000008  09 0a        ..\n"
    assert hexdump == hexdump_truth



import sdr


def test_degree_divides_4():
    assert sdr.is_preferred_pair("x^4 + x + 1", "x^4 + x^3 + 1") is False


def test_degree_3():
    assert sdr.is_preferred_pair("x^3 + x + 1", "x^3 + x^2 + 1") is True


def test_degree_5():
    assert sdr.is_preferred_pair("x^5 + x^2 + 1", "x^5 + x^3 + x^2 + x + 1") is True
    assert sdr.is_preferred_pair("x^5 + x^2 + 1", "x^5 + x^3 + 1") is False



import numpy as np
import scipy.stats

import sdr


def test_normal_normal():
    X = scipy.stats.norm(loc=3, scale=0.5)
    Y = scipy.stats.norm(loc=5, scale=1.5)
    _verify(X, Y)


def test_rayleigh_rayleigh():
    X = scipy.stats.rayleigh(scale=1)
    Y = scipy.stats.rayleigh(loc=1, scale=2)
    _verify(X, Y)


def test_rician_rician():
    X = scipy.stats.rice(2)
    Y = scipy.stats.rice(3)
    _verify(X, Y)


def test_normal_rayleigh():
    X = scipy.stats.norm(loc=-1, scale=0.5)
    Y = scipy.stats.rayleigh(loc=2, scale=1.5)
    _verify(X, Y)


def test_rayleigh_rician():
    X = scipy.stats.rayleigh(scale=1)
    Y = scipy.stats.rice(3)
    _verify(X, Y)


def _verify(X, Y):
    # Empirically compute the distribution
    z = X.rvs(250_000) * Y.rvs(250_000)
    hist, bins = np.histogram(z, bins=51, density=True)
    x = bins[1:] - np.diff(bins) / 2

    # Numerically compute the distribution, only do so over the histogram bins (for speed)
    Z = sdr.multiply_distributions(X, Y, x)

    if False:
        import matplotlib.pyplot as plt

        plt.figure()
        plt.plot(x, X.pdf(x), label="X")
        plt.plot(x, Y.pdf(x), label="Y")
        plt.plot(x, Z.pdf(x), label="X * Y")
        plt.hist(z, bins=51, cumulative=False, density=True, histtype="step", label="X * Y empirical")
        plt.legend()
        plt.xlabel("Random variable")
        plt.ylabel("Probability density")
        plt.title("Product of two distributions")
        plt.show()

    assert np.allclose(Z.pdf(x), hist, atol=np.max(hist) * 0.2)



"""
A module for fractional delay filters.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.signal

from .._helper import export, verify_scalar
from ._design_fir import _normalize_passband
from ._fir import FIR


def _ideal_fractional_delay(length: int, delay: float) -> npt.NDArray[np.float64]:
    """
    Returns the ideal fractional delay filter impulse response.
    """
    n = np.arange(-length // 2 + 1, length // 2 + 1)  # Sample indices
    h_ideal = np.sinc(n - delay)  # Ideal filter impulse response
    return h_ideal


@export
def fractional_delay_fir(
    length: int,
    delay: float,
) -> npt.NDArray[np.float64]:
    r"""
    Designs a fractional delay FIR filter impulse response $h[n]$ using the Kaiser window method.

    Arguments:
        length: The filter length $L$. Filters with even length have best performance.
            Filters with odd length are equivalent to an even-length filter with an appended zero.
        delay: The fractional delay $0 \le \Delta n \le 1$.

    Returns:
        The filter impulse response $h[n]$ with length $L$. The center of the passband has 0 dB gain.

    Notes:
        The filter group delay is $\tau = L_{even}/2 - 1 + \Delta n$ at DC.

    References:
        - https://www.mathworks.com/help/dsp/ref/designfracdelayfir.html

    Examples:
        Design a $\Delta n = 0.25$ delay filter with length 8. Observe the width and flatness of the frequency
        response passband. Also observe the group delay of 3.25 at DC.

        .. ipython:: python

            h_8 = sdr.fractional_delay_fir(8, 0.25)

            @savefig sdr_fractional_delay_fir_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_8);

            @savefig sdr_fractional_delay_fir_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_8); \
            plt.ylim(-4, 1);

            @savefig sdr_fractional_delay_fir_3.png
            plt.figure(); \
            sdr.plot.group_delay(h_8);

        Compare the magnitude response and group delay of filters with different lengths.

        .. ipython:: python

            h_16 = sdr.fractional_delay_fir(16, 0.25); \
            h_32 = sdr.fractional_delay_fir(32, 0.25); \
            h_64 = sdr.fractional_delay_fir(64, 0.25)

            @savefig sdr_fractional_delay_fir_4.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_8, label="Length 8"); \
            sdr.plot.magnitude_response(h_16, label="Length 16"); \
            sdr.plot.magnitude_response(h_32, label="Length 32"); \
            sdr.plot.magnitude_response(h_64, label="Length 64"); \
            plt.ylim(-4, 1);

            @savefig sdr_fractional_delay_fir_5.png
            plt.figure(); \
            sdr.plot.group_delay(h_8, label="Length 8"); \
            sdr.plot.group_delay(h_16, label="Length 16"); \
            sdr.plot.group_delay(h_32, label="Length 32"); \
            sdr.plot.group_delay(h_64, label="Length 64");

    Group:
        dsp-arbitrary-resampling
    """
    verify_scalar(length, int=True, inclusive_min=2)
    verify_scalar(delay, float=True, inclusive_min=0, inclusive_max=1)

    N = length - (length % 2)  # The length guaranteed to be even
    h_ideal = _ideal_fractional_delay(N, delay)

    if N == 2:
        beta = 0
    elif N == 4:
        beta = 2.21
    else:
        beta = (11.01299 * N**2 + 2395.00455 * N - 6226.46055) / (1.00000 * N**2 + 326.73886 * N + 1094.40241)
    h_window = scipy.signal.windows.kaiser(N, beta=beta)

    h = h_ideal * h_window

    if N < length:
        h = np.pad(h, (0, length - N), mode="constant")

    h = _normalize_passband(h, 0)

    return h


@export
class FractionalDelay(FIR):
    r"""
    Implements a fractional delay FIR filter.

    Examples:
        Design $\Delta n = 0.21719$ delay filters with various lengths.
        Examine the width and flatness of the frequency response passband.

        .. ipython:: python

            plt.figure();
            for length in [4, 8, 16, 32, 64, 128]:
                fir = sdr.FractionalDelay(length, 0.21719)
                sdr.plot.magnitude_response(fir, label=f"$L = {length}$")
            @savefig sdr_FractionalDelay_1.png
            plt.legend(loc="lower left");

        Design filters with length $L = 8$ and various fractional delays.
        Examine the effects on the magnitude response outside the passband as a function of the fractional delay.
        Note the symmetry about $\Delta n = 0.5$ and that the out-of-band magnitude response is worst at
        $\Delta n = 0.5$.

        .. ipython:: python

            plt.figure();
            for delay in np.arange(0.1, 1, 0.1):
                fir = sdr.FractionalDelay(8, delay)
                sdr.plot.magnitude_response(fir, label=f"$\Delta n = {delay:0.1f}$")
            @savefig sdr_FractionalDelay_2.png
            plt.ylim(-10, 1); \
            plt.legend(loc="lower left");

        Examine the effects on the group delay outside the passband as a function of the fractional delay.
        Note the symmetry about $\Delta n = 0.5$. The out-of-band group delay is worst around $\Delta n \approx 0.5$,
        however the group delay is perfectly flat at exactly $\Delta n = 0.5$.

        .. ipython:: python

            plt.figure();
            for delay in np.arange(0.1, 1, 0.1):
                fir = sdr.FractionalDelay(8, delay)
                sdr.plot.group_delay(fir, label=f"$\Delta n = {delay:0.1f}$")
            @savefig sdr_FractionalDelay_3.png
            plt.legend(loc="lower left");

    Group:
        dsp-arbitrary-resampling
    """

    def __init__(
        self,
        length: int,
        delay: float,
    ):
        r"""
        Creates a fractional delay FIR filter.

        Arguments:
            length: The filter length $L$. Filters with even length have best performance.
                Filters with odd length are equivalent to an even-length filter with an appended zero.
            delay: The fractional delay $0 \le \Delta n \le 1$.
        """
        verify_scalar(length, int=True, inclusive_min=2)
        verify_scalar(delay, float=True, inclusive_min=0, inclusive_max=1)

        h = fractional_delay_fir(length, delay)
        super().__init__(h)

        self._delay = length // 2 - 1 + delay



"""
A module for numerically controlled oscillators (NCO).
"""

from __future__ import annotations

from typing import Any, overload

import numpy as np
import numpy.typing as npt
from typing_extensions import Literal

from ._helper import (
    convert_output,
    export,
    verify_arraylike,
    verify_at_least_one_specified,
    verify_literal,
    verify_same_shape,
    verify_scalar,
)


@export
class NCO:
    r"""
    Implements a numerically controlled oscillator (NCO).

    Notes:
        .. code-block:: text
           :caption: Numerically Controlled Oscillator Block Diagram

                          constant           constant
                         increment            offset    p[n]
                             |                  |         |
                    +----+   v   +------+       v         v   +-----------+
            f[n] -->| K0 |-->@-->| z^-1 |---+-->@-------->@-->| output(.) |--> y[n]
                    +----+   ^   +------+   |                 +-----------+
                             |              |
                             +--------------+

            f[n] = Input frequency signal (radians/sample)
            p[n] = Input phase signal (radians)
            y[n] = Output complex signal
            K0 = NCO gain
            increment = Constant phase accumulation (radians/sample)
            offset = Absolute phase offset (radians)
            output(.) = Phase-to-output mapping function, either: ., sin(.), cos(.), exp(j .)
            z^-1 = Unit delay
            @ = Adder

    Examples:
        Create an NCO with a constant phase increment of $2 \pi / 20$ radians/sample and a constant phase offset
        of $\pi$ radians.

        .. ipython:: python

            nco = sdr.NCO(increment=2 * np.pi / 20, offset=np.pi); \
            y = nco.step(100)

            @savefig sdr_NCO_1.png
            plt.figure(); \
            sdr.plot.time_domain(y, marker="."); \
            plt.title("Constant frequency NCO");

        Create an NCO with a constant phase increment of 0 radians/sample and a constant phase offset
        of 0 radians. Then step the NCO with a FSK frequency signal.

        .. ipython:: python

            nco = sdr.NCO(); \
            freq = np.array([1, 2, -3, 0, 2, -1]) * 2 * np.pi / 60; \
            freq = np.repeat(freq, 20); \
            y = nco(freq=freq)

            @savefig sdr_NCO_2.png
            plt.figure(); \
            sdr.plot.time_domain(y, marker="."); \
            plt.title("NCO implementing CP-FSK modulation");

        Create an NCO with a constant phase increment of $2 \pi / 57$ radians/sample and a constant phase offset
        of 0 radians. Then step the NCO with a BPSK phase signal.

        .. ipython:: python

            nco = sdr.NCO(increment=2 * np.pi / 57); \
            phase = np.array([0, 1, 0, 1, 0, 1]) * np.pi; \
            phase = np.repeat(phase, 20); \
            y = nco(phase=phase)

            @savefig sdr_NCO_3.png
            plt.figure(); \
            sdr.plot.time_domain(y, marker="."); \
            plt.title("NCO implementing BPSK modulation");

        See the :ref:`phase-locked-loop` example.

    Group:
        pll
    """

    def __init__(self, gain: float = 1.0, increment: float = 0.0, offset: float = 0.0):
        r"""
        Creates a numerically controlled oscillator (NCO).

        Arguments:
            gain: The NCO gain $K_0$.
            increment: The constant accumulation $\omega$ of the NCO in radians/sample.
            offset: The absolute offset $\theta$ of the NCO in radians.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        self._K0 = verify_scalar(gain, float=True, positive=True)
        self._increment = verify_scalar(increment, float=True)
        self._offset = verify_scalar(offset, float=True)
        self._z_prev: float  # Will be updated in reset()
        self.reset()

    def reset(self):
        r"""
        Resets the NCO.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        self._z_prev = 0.0

    @overload
    def __call__(
        self,
        freq: npt.NDArray | None = None,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
        phase: npt.NDArray | None = None,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
        output: Literal["phase", "sine", "cosine"] = "complex-exp",
    ) -> npt.NDArray[np.float64]: ...

    @overload
    def __call__(
        self,
        freq: npt.NDArray | None = None,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
        phase: npt.NDArray | None = None,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
        output: Literal["complex-exp"] = "complex-exp",
    ) -> npt.NDArray[np.complex128]: ...

    def __call__(
        self,
        freq: Any = None,
        phase: Any = None,
        output: Any = "complex-exp",
    ) -> Any:
        """
        Steps the NCO with variable frequency and/or phase signals.

        Arguments:
            freq: The variable frequency signal $f[n]$ in radians/sample. This input signal varies the per-sample
                phase increment of the NCO. If `None`, the signal is all zeros.
            phase: The variable phase signal $p[n]$ in radians. This input signal varies the per-sample phase offset
                of the NCO. If `None`, the signal is all zeros.
            output: The format of the output signal $y[n]$. Options are the accumulated phase, sine, cosine, or
                complex exponential.

        Returns:
            The output signal $y[n]$.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        verify_at_least_one_specified(freq, phase)
        freq = verify_arraylike(freq, optional=True, float=True, atleast_1d=True, ndim=1)
        phase = verify_arraylike(phase, optional=True, float=True, atleast_1d=True, ndim=1)
        if freq is None:
            freq = np.zeros_like(phase)
        if phase is None:
            phase = np.zeros_like(freq)
        verify_same_shape(freq, phase)
        verify_literal(output, ["phase", "sine", "cosine", "complex-exp"])

        # Scale the input by the NCO gain and add the constant accumulation to every sample
        z = freq * self.gain + self.increment

        # Insert the previous output in front of the current input. Then run a cumulative sum over all samples.
        z = np.insert(z, 0, self._z_prev)
        z = np.cumsum(z)
        z, self._z_prev = z[0:-1], z[-1]

        # Add the absolute offset to every sample
        y = z + self.offset + phase

        if output == "phase":
            pass
        elif output == "complex-exp":
            y = np.exp(1j * y)
        elif output == "sine":
            y = np.sin(y)
        elif output == "cosine":
            y = np.cos(y)

        return convert_output(y, squeeze=True)

    def step(self, N: int) -> npt.NDArray[np.complex128]:
        """
        Steps the NCO forward by $N$ samples.

        Arguments:
            N: The number of samples $N$ to step the NCO forward.

        Returns:
            The output complex signal $y[n]$.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        verify_scalar(N, int=True, positive=True)

        x = np.zeros(N)
        y = self(x)

        return y

    @property
    def gain(self) -> float:
        """
        (Settable) The NCO gain $K_0$.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._K0

    @gain.setter
    def gain(self, value: float):
        self._K0 = value

    @property
    def increment(self) -> float:
        r"""
        (Settable) The constant phase accumulation $\omega$ of the NCO in radians/sample.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._increment

    @increment.setter
    def increment(self, value: float):
        self._increment = value

    @property
    def offset(self) -> float:
        r"""
        (Settable) The absolute phase offset $\theta$ of the NCO in radians.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._offset

    @offset.setter
    def offset(self, value: float):
        self._offset = value



"""
A module for finite impulse response (FIR) filters.
"""

from __future__ import annotations

from typing import Any, overload

import numpy as np
import numpy.typing as npt
import scipy.integrate
import scipy.signal
from typing_extensions import Literal

from .._helper import convert_output, export, verify_arraylike, verify_bool, verify_literal, verify_scalar
from ._common import frequency_response


@export
class FIR:
    r"""
    Implements a finite impulse response (FIR) filter.

    This class is a wrapper for the :func:`scipy.signal.convolve` function. It supports one-time filtering
    and streamed filtering.

    Notes:
        A FIR filter is defined by its feedforward coefficients $h[n]$.

        $$y[n] = \sum_{i=0}^{N} h[i] \cdot x[n-i] .$$

        The transfer function of the filter is

        $$H(z) = \sum\limits_{i=0}^{N} h[i] \cdot z^{-i} .$$

        .. code-block:: text
            :caption: FIR Block Diagram

                      +------+    +------+            +------+    +------+
            x[n] --+->| z^-1 |-+->| z^-1 |-+--...--+->| z^-1 |-+->| z^-1 |-+
                   |  +------+ |  +------+ |       |  +------+ |  +------+ |
                   |           |           |       |           |           |
              h[0] |      h[1] |      h[2] |       |    h[N-1] |      h[N] |
                   |           |           |       |           |           |
                   +---------->@---------->@--...->@---------->@---------->@---> y[n]

    Examples:
        See the :ref:`fir-filters` example.

    Group:
        dsp-fir-filtering
    """

    def __init__(
        self,
        h: npt.ArrayLike,
        streaming: bool = False,
    ):
        """
        Creates an FIR filter.

        Arguments:
            h: The feedforward coefficients $h[n]$.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~FIR.__call__()`.

        Examples:
            See the :ref:`fir-filters` example.
        """
        self._taps = verify_arraylike(h, ndim=1)
        self._streaming = verify_bool(streaming)

        self._delay = self.taps.size // 2
        self._state: npt.NDArray  # The filter state. Will be updated in reset().
        self.reset()

    ##############################################################################
    # Special methods
    ##############################################################################

    def __call__(
        self,
        x: npt.ArrayLike,
        mode: Literal["full", "valid", "same"] = "full",
    ) -> npt.NDArray:
        r"""
        Filters the input signal $x[n]$ with the FIR filter.

        Arguments:
            x: The input signal $x[n]$ with length $L$.
            mode: The non-streaming convolution mode.

                - `"same"`: The output signal $y[n]$ has length $L$. Output sample 0 aligns with input sample 0.
                - `"full"`: The full convolution is performed. The output signal $y[n]$ has length $L + N$,
                  where $N$ is the order of the filter. Output sample :obj:`~sdr.FIR.delay` aligns
                  with input sample 0.

                In streaming mode, the `"full"` convolution is performed. However, for each $L$ input samples
                only $L$ output samples are produced per call. A final call to :meth:`~sdr.FIR.flush()`
                is required to flush the filter state.

        Returns:
            The filtered signal $y[n]$. The output length is dictated by the `mode` argument.

        Examples:
            See the :ref:`fir-filters` example.
        """
        x = verify_arraylike(x, atleast_1d=True, ndim=1)
        verify_literal(mode, ["full", "valid", "same"])

        if self.streaming:
            # Prepend previous inputs from last __call__() call
            x_pad = np.concatenate((self._state, x))
            y = scipy.signal.convolve(x_pad, self.taps, mode="valid")
            self._state = x_pad[-(self.taps.size - 1) :]
        else:
            y = scipy.signal.convolve(x, self.taps, mode=mode)

        return convert_output(y)

    def __repr__(self) -> str:
        return f"sdr.{type(self).__name__}({self.taps.tolist()}, streaming={self.streaming})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  taps: {self.taps.shape} shape"
        string += f"\n    {self.taps.tolist()}"
        string += f"\n  delay: {self.delay}"
        string += f"\n  streaming: {self.streaming}"
        return string

    def __len__(self) -> int:
        """
        Returns the filter length $N + 1$.

        Examples:
            See the :ref:`fir-filters` example.
        """
        return self.taps.size

    ##############################################################################
    # Streaming mode
    ##############################################################################

    def reset(self):
        """
        Resets the filter state. Only useful when using streaming mode.

        Examples:
            See the :ref:`fir-filters` example.

        Group:
            Streaming mode only
        """
        self._state = np.zeros(self.taps.size - 1, dtype=self.taps.dtype)

    def flush(self) -> npt.NDArray:
        """
        Flushes the filter state by passing zeros through the filter. Only useful when using streaming mode.

        Returns:
            The remaining filtered signal $y[n]$.

        Examples:
            See the :ref:`fir-filters` example.

        Group:
            Streaming mode only
        """
        x = np.zeros_like(self.state)
        y = self(x)
        return y

    @property
    def streaming(self) -> bool:
        """
        Indicates whether the filter is in streaming mode.

        In streaming mode, the filter state is preserved between calls to :meth:`~FIR.__call__()`.

        Examples:
            See the :ref:`fir-filters` example.

        Group:
            Streaming mode only
        """
        return self._streaming

    @property
    def state(self) -> npt.NDArray:
        """
        The filter state consisting of the previous $N$ inputs.

        Examples:
            See the :ref:`fir-filters` example.

        Group:
            Streaming mode only
        """
        return self._state

    ##############################################################################
    # Methods
    ##############################################################################

    def impulse_response(self, N: int | None = None) -> npt.NDArray:
        r"""
        Returns the impulse response $h[n]$ of the FIR filter.

        The impulse response $h[n]$ is the filter output when the input is an impulse $\delta[n]$.

        Arguments:
            N: The number of samples to return. The default is the filter length.

        Returns:
            The impulse response of the IIR filter $h[n]$.

        See Also:
            sdr.plot.impulse_response

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(N, optional=True, int=True, inclusive_min=self.taps.size)
        if N is None:
            N = self.taps.size

        # Delta impulse function
        d = np.zeros(N - self.taps.size + 1, dtype=float)
        d[0] = 1

        h = scipy.signal.convolve(d, self.taps, mode="full")

        return h

    def step_response(self, N: int | None = None) -> npt.NDArray:
        """
        Returns the step response $s[n]$ of the FIR filter.

        The step response $s[n]$ is the filter output when the input is a unit step $u[n]$.

        Arguments:
            N: The number of samples to return. The default is the filter length.

        Returns:
            The step response of the FIR filter $s[n]$.

        See Also:
            sdr.plot.step_response

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(N, optional=True, int=True, inclusive_min=self.taps.size)
        if N is None:
            N = self.taps.size

        # Unit step function
        u = np.ones(N - self.taps.size + 1, dtype=float)

        s = scipy.signal.convolve(u, self.taps, mode="full")

        return s

    @overload
    def frequency_response(
        self,
        freqs: int = 1024,
        sample_rate: float = 1.0,
        whole: bool = True,
        decades: int | None = None,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]: ...

    @overload
    def frequency_response(
        self,
        freqs: float,
        sample_rate: float = 1.0,
    ) -> complex: ...

    @overload
    def frequency_response(
        self,
        freqs: npt.NDArray[np.float64],
        sample_rate: float = 1.0,
    ) -> npt.NDArray[np.complex128]: ...

    def frequency_response(
        self,
        freqs: Any = 1024,
        sample_rate: Any = 1.0,
        whole: Any = True,
        decades: Any | None = None,
    ) -> Any:
        r"""
        Returns the frequency response $H(\omega)$ of the FIR filter.

        Arguments:
            freqs: The frequency specification.

                - `int`: The number of frequency points. The endpoint is not included.
                - `float`: A single frequency.
                - `npt.NDArray[float]`: Multiple frequencies.

            sample_rate: The sample rate $f_s$ of the filter in samples/s.
            whole: Only used if `freqs` is an integer.

                - `True`: The maximum frequency is `max_f = sample_rate`.
                - `False`: The maximum frequency is `max_f = sample_rate / 2`.

            decades: Only used if `freqs` is an integer.

                - `None`: `f = np.linspace(0, max_f, freqs, endpoint=False)`.
                - `int`: `f = np.logspace(np.log10(max_f) - decades), np.log10(max_f), freqs, endpoint=False)`.

        Returns:
            - The frequency vector $f$, only if `freqs` is an integer.
            - The frequency response of the FIR filter $H(\omega)$.

        See Also:
            sdr.plot.magnitude_response, sdr.plot.phase_response

        Examples:
            .. ipython:: python

                h = sdr.lowpass_fir(100, 0.2, window="hamming"); \
                fir = sdr.FIR(h)

            Compute the frequency response at 1024 evenly spaced frequencies.

            .. ipython:: python

                fir.frequency_response()

            Compute the frequency response at 0.0 rad/s.

            .. ipython:: python

                fir.frequency_response(0.0)

            Compute the frequency response at several frequencies in Hz.

            .. ipython:: python

                fir.frequency_response([100, 200, 300, 400], sample_rate=1000)
        """
        verify_scalar(sample_rate, float=True, positive=True)
        verify_bool(whole)
        verify_scalar(decades, optional=True, int=True, positive=True)

        f, H = frequency_response(self.taps, 1, freqs, sample_rate, whole, decades)
        if isinstance(freqs, int):
            return f, H
        elif isinstance(freqs, float):
            return H[0]
        else:
            return H

    def group_delay(self, sample_rate: float = 1.0, N: int = 1024) -> tuple[npt.NDArray, npt.NDArray]:
        r"""
        Returns the group delay $\tau_g(\omega)$ of the FIR filter.

        Arguments:
            sample_rate: The sample rate $f_s$ of the filter in samples/s.
            N: The number of samples in the group delay.

        Returns:
            - The frequencies $f$ from $-f_s/2$ to $f_s/2$ in Hz.
            - The group delay of the FIR filter $\tau_g(\omega)$.

        See Also:
            sdr.plot.group_delay

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(sample_rate, float=True, positive=True)
        verify_scalar(N, int=True, positive=True)

        f, gd = scipy.signal.group_delay((self.taps, 1), w=N, whole=True, fs=sample_rate)

        f[f >= 0.5 * sample_rate] -= sample_rate  # Wrap frequencies from [0, 1) to [-0.5, 0.5)
        f = np.fft.fftshift(f)
        gd = np.fft.fftshift(gd)

        return f, gd

    def phase_delay(self, sample_rate: float = 1.0, N: int = 1024) -> tuple[npt.NDArray, npt.NDArray]:
        r"""
        Returns the phase delay $\tau_{\phi}(\omega)$ of the FIR filter.

        Arguments:
            sample_rate: The sample rate $f_s$ of the filter in samples/s.
            N: The number of samples in the phase delay.

        Returns:
            - The frequencies $f$ from $-f_s/2$ to $f_s/2$ in Hz.
            - The phase delay of the FIR filter $\tau_{\phi}(\omega)$.

        See Also:
            sdr.plot.phase_delay

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(sample_rate, float=True, positive=True)
        verify_scalar(N, int=True, positive=True)

        f, H = scipy.signal.freqz(self.taps, 1, worN=N, whole=True, fs=sample_rate)

        f -= sample_rate / 2
        H = np.fft.fftshift(H)

        theta = np.unwrap(np.angle(H), period=np.pi)
        theta -= theta[np.argmin(np.abs(f))]  # Set omega=0 to have phase of 0
        tau_phi = -theta / (2 * np.pi * f)
        tau_phi[np.argmin(np.abs(f))] = np.nan  # Avoid crazy result when dividing by near zero

        return f, tau_phi

    def noise_bandwidth(self, sample_rate: float = 1.0) -> float:
        r"""
        Returns the noise bandwidth $B_n$ of the FIR filter.

        Arguments:
            sample_rate: The sample rate $f_s$ of the filter in samples/s.

        Returns:
            The noise bandwidth of the FIR filter $B_n$ in Hz.

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(sample_rate, float=True, positive=True)

        # Compute the frequency response
        f, H = scipy.signal.freqz(self.taps, 1, worN=2**20, whole=True, fs=sample_rate)
        H = np.abs(H) ** 2

        # Compute the noise bandwidth
        B_n = scipy.integrate.simpson(y=H, x=f) / np.max(H)

        return B_n

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def taps(self) -> npt.NDArray:
        """
        The feedforward taps $h[n]$ with length $N + 1$.

        Examples:
            See the :ref:`fir-filters` example.
        """
        return self._taps

    @property
    def order(self) -> int:
        """
        The order of the FIR filter $N$.

        Examples:
            See the :ref:`fir-filters` example.
        """
        return self.taps.size - 1

    @property
    def delay(self) -> int:
        r"""
        The delay of the FIR filter $d = \lfloor \frac{N + 1}{2} \rfloor$ in samples.

        Examples:
            See the :ref:`fir-filters` example.
        """
        return self._delay



import numpy as np

import sdr


def test_decode_1():
    """
    MATLAB:
        >> diff_dec = comm.DifferentialDecoder;
        >> y = randi([0 1], 10, 1); y
        >> diff_dec(y)
    """
    y = [1, 0, 1, 1, 0, 0, 0, 0, 0, 1]
    x = sdr.diff_decode(y)
    x_truth = [1, 1, 1, 0, 1, 0, 0, 0, 0, 1]
    assert np.array_equal(x, x_truth)


# TODO: Add more MATLAB tests



import numpy as np
import pytest

import sdr


def test_types():
    y = sdr.bsc(0, 0.5)
    assert isinstance(y, int)

    y = sdr.bsc([0, 1], 0.5)
    assert isinstance(y, np.ndarray)


def test_bit_flips():
    rng = np.random.default_rng()
    p = rng.uniform(0.2, 0.8)
    N = int(1000 / p)
    x = rng.integers(0, 2, N)
    y = sdr.bsc(x, p)
    assert np.count_nonzero(x != y) / N == pytest.approx(p, rel=1e-1)



"""
A Python package for software-defined radio (SDR) applications.
"""

try:
    from ._version import __version__, __version_tuple__
except ModuleNotFoundError:  # pragma: no cover
    import warnings

    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)
    warnings.warn(
        "An error occurred during package install where setuptools_scm failed to create a _version.py file. "
        "Defaulting version to 0.0.0.",
        stacklevel=3,
    )

from . import plot
from ._coding import *
from ._conversion import *
from ._data import *
from ._detection import *
from ._estimation import *
from ._farrow import *
from ._filter import *
from ._link_budget import *
from ._measurement import *
from ._modulation import *
from ._nco import *
from ._probability import *
from ._sequence import *
from ._signal import *
from ._simulation import *
from ._synchronization import *



import galois

import sdr


def test_degree_divides_4():
    assert list(sdr.preferred_pairs(4)) == []
    assert list(sdr.preferred_pairs(8)) == []
    assert list(sdr.preferred_pairs(12)) == []


def test_degree_3():
    assert list(sdr.preferred_pairs(3)) == [
        (galois.Poly.Str("x^3 + x + 1"), galois.Poly.Str("x^3 + x^2 + 1")),
    ]


def test_degree_5():
    assert list(sdr.preferred_pairs(5)) == [
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^3 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^4 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x + 1")),
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x^2 + 1")),
        (galois.Poly.Str("x^5 + x^3 + 1"), galois.Poly.Str("x^5 + x^3 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^3 + 1"), galois.Poly.Str("x^5 + x^4 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^3 + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x + 1")),
        (galois.Poly.Str("x^5 + x^3 + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x^2 + 1")),
        (galois.Poly.Str("x^5 + x^3 + x^2 + x + 1"), galois.Poly.Str("x^5 + x^4 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^3 + x^2 + x + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x + 1")),
        (galois.Poly.Str("x^5 + x^4 + x^2 + x + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x^2 + 1")),
        (galois.Poly.Str("x^5 + x^4 + x^3 + x + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x^2 + 1")),
    ]


def test_degree_5_specific():
    assert list(sdr.preferred_pairs(5, poly="x^5 + x^2 + 1")) == [
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^3 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^4 + x^2 + x + 1")),
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x + 1")),
        (galois.Poly.Str("x^5 + x^2 + 1"), galois.Poly.Str("x^5 + x^4 + x^3 + x^2 + 1")),
    ]



import numpy as np

import sdr


def test_bpsk_rect():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.rectangular(sps)
    psk = sdr.PSK(2, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_demodulation(psk)


def test_bpsk_srrc():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.root_raised_cosine(0.5, 16, sps)
    psk = sdr.PSK(2, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_demodulation(psk)


def test_qpsk_rect():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.rectangular(sps)
    psk = sdr.PSK(4, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_demodulation(psk)


def test_qpsk_srrc():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.root_raised_cosine(0.5, 16, sps)
    psk = sdr.PSK(4, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_demodulation(psk)


def test_8psk_rect():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.rectangular(sps)
    psk = sdr.PSK(8, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_demodulation(psk)


def test_8psk_srrc():
    rng = np.random.default_rng()
    phi = rng.uniform(0, 360)
    sps = 10
    h = sdr.root_raised_cosine(0.5, 16, sps)
    psk = sdr.PSK(8, phase_offset=phi, sps=sps, pulse_shape=h)
    _verify_demodulation(psk)


def _verify_demodulation(psk: sdr.PSK):
    rng = np.random.default_rng()
    s = rng.integers(0, psk.order, 20)
    a = psk.map_symbols(s)
    x = psk.modulate(s)
    s_hat, a_tilde, a_hat = psk.demodulate(x)

    assert np.array_equal(s, s_hat)
    np.testing.assert_array_almost_equal(a, a_hat, decimal=3)



"""
A module containing various spectral estimation techniques.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.signal
from typing_extensions import Literal

from .._conversion import db
from .._helper import export, verify_arraylike, verify_isinstance, verify_literal, verify_scalar
from ._helper import integer_x_axis, verify_sample_rate
from ._rc_params import RC_PARAMS
from ._units import freq_units, time_units


@export
def periodogram(
    x: npt.ArrayLike,
    sample_rate: float | None = None,
    window: str | float | tuple | None = "hann",
    length: int | None = None,
    overlap: int | None = None,
    fft: int | None = None,
    detrend: Literal["constant", "linear", False] = False,
    average: Literal["mean", "median"] = "mean",
    ax: plt.Axes | None = None,
    x_axis: Literal["auto", "one-sided", "two-sided", "log"] = "auto",
    y_axis: Literal["linear", "log"] = "log",
    **kwargs,
):
    r"""
    Plots the estimated power spectral density $P_{xx}$ of a time-domain signal $x[n]$ using Welch's method.

    Note:
        This function uses :func:`scipy.signal.welch()` to estimate the power spectral density of the
        time-domain signal.

    Arguments:
        x: The time-domain signal $x[n]$.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.
        length: The length of each segment in samples. If `None`, the length is set to 256.
        overlap: The number of samples to overlap between segments. If `None`, the overlap is set to `length // 2`.
        fft: The number of points to use in the FFT. If `None`, the FFT length is set to `length`.
        detrend: The type of detrending to apply. Options are to remove the mean or a linear trend from each segment.
        average: The type of averaging to use. Options are to average the periodograms using the mean or median.
        ax: The axis to plot on. If `None`, the current axis is used.
        x_axis: The x-axis scaling. Options are to display a one-sided spectrum, a two-sided spectrum, or
            one-sided spectrum with a logarithmic frequency axis. The default is `"auto"` which selects `"one-sided"`
            for real-valued signals and `"two-sided"` for complex-valued signals.
        y_axis: The y-axis scaling. Options are to display a linear or logarithmic power spectral density.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Group:
        plot-spectral-estimation
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    # verify_isinstance(window, str, optional=True)
    verify_scalar(length, optional=True, int=True, positive=True)
    verify_scalar(overlap, optional=True, int=True, positive=True)
    verify_scalar(fft, optional=True, int=True, positive=True)
    verify_literal(detrend, ["constant", "linear", False])
    verify_literal(average, ["mean", "median"])
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(x_axis, ["auto", "one-sided", "two-sided", "log"])
    verify_literal(y_axis, ["linear", "log"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if x_axis == "auto":
            x_axis = "one-sided" if np.isrealobj(x) else "two-sided"

        f, Pxx = scipy.signal.welch(
            x,
            fs=sample_rate,
            window=window,
            nperseg=length,
            noverlap=overlap,
            nfft=fft,
            detrend=detrend,
            return_onesided=x_axis != "two-sided",
            average=average,
        )

        if y_axis == "log":
            Pxx = db(Pxx)

        if x_axis == "two-sided":
            f[f >= 0.5 * sample_rate] -= sample_rate  # Wrap frequencies from [0, 1) to [-0.5, 0.5)
            f = np.fft.fftshift(f)
            Pxx = np.fft.fftshift(Pxx)

        if sample_rate_provided:
            units, scalar = freq_units(f)
            f *= scalar

        if x_axis == "log":
            ax.semilogx(f, Pxx, **kwargs)
        else:
            ax.plot(f, Pxx, **kwargs)

        ax.grid(True, which="both")
        if "label" in kwargs:
            ax.legend()

        if sample_rate_provided:
            ax.set_xlabel(f"Frequency ({units}), $f$")
        else:
            ax.set_xlabel("Normalized frequency, $f / f_s$")

        if y_axis == "log":
            ax.set_ylabel("Power density (dB/Hz)")
        else:
            ax.set_ylabel("Power density (W/Hz)")

        ax.set_title("Power spectral density")


@export
def spectrogram(
    x: npt.ArrayLike,
    sample_rate: float | None = None,
    window: str | float | tuple | None = "hann",
    length: int | None = None,
    overlap: int | None = None,
    fft: int | None = None,
    detrend: Literal["constant", "linear", False] = False,
    ax: plt.Axes | None = None,
    y_axis: Literal["auto", "one-sided", "two-sided"] = "auto",
    # persistence: bool = False,
    # colorbar: bool = True,
    **kwargs,
):
    r"""
    Plots the spectrogram of a time-domain signal $x[n]$ using Welch's method.

    Note:
        This function uses :func:`scipy.signal.spectrogram` to estimate the spectrogram of the time-domain signal.

    Arguments:
        x: The time-domain signal $x[n]$.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be label as "Samples" and the y-axis as "Normalized frequency".
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.
        length: The length of each segment in samples. If `None`, the length is set to 256.
        overlap: The number of samples to overlap between segments. If `None`, the overlap is set to `length // 2`.
        fft: The number of points to use in the FFT. If `None`, the FFT length is set to `length`.
        detrend: The type of detrending to apply. Options are to remove the mean or a linear trend from each segment.
        ax: The axis to plot on. If `None`, the current axis is used.
        y_axis: The y-axis scaling. Options are to display a one-sided spectrum or a two-sided spectrum.
            The default is `"auto"` which selects `"one-sided"` for real-valued signals and `"two-sided"` for
            complex-valued signals.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.pcolormesh`.
            The following keyword arguments are set by default. The defaults may be overwritten.

            - `"vmin"`: 10th percentile
            - `"vmax"`: 100th percentile
            - `"shading"`: `"gouraud"`

    Group:
        plot-spectral-estimation
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    # verify_isinstance(window, str, optional=True)
    verify_scalar(length, optional=True, int=True, positive=True)
    verify_scalar(overlap, optional=True, int=True, positive=True)
    verify_scalar(fft, optional=True, int=True, positive=True)
    verify_literal(detrend, ["constant", "linear", False])
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(y_axis, ["auto", "one-sided", "two-sided"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if y_axis == "auto":
            y_axis = "one-sided" if np.isrealobj(x) else "two-sided"

        f, t, Sxx = scipy.signal.spectrogram(
            x,
            fs=sample_rate,
            window=window,
            nperseg=length,
            noverlap=overlap,
            nfft=fft,
            detrend=detrend,
            return_onesided=y_axis != "two-sided",
            mode="psd",
        )
        Sxx = db(Sxx)

        if y_axis == "one-sided" and np.iscomplexobj(x):
            # If complex data, the spectrogram always returns a two-sided spectrum. So we need to remove the second half.
            f = f[0 : f.size // 2]
            Sxx = Sxx[0 : Sxx.shape[0] // 2, :]
        if y_axis == "two-sided":
            f = np.fft.fftshift(f)
            Sxx = np.fft.fftshift(Sxx, axes=0)

        if sample_rate_provided:
            f_units, scalar = freq_units(f)
            f *= scalar
            t_units, scalar = time_units(t)
            t *= scalar

        default_kwargs = {
            "vmin": np.percentile(Sxx, 10),
            "vmax": np.percentile(Sxx, 100),
            "shading": "gouraud",
        }
        kwargs = {**default_kwargs, **kwargs}

        ax.pcolormesh(t, f, Sxx, **kwargs)
        if sample_rate_provided:
            ax.set_xlabel(f"Time ({t_units}), $t$")
            ax.set_ylabel(f"Frequency ({f_units}), $f$")
        else:
            integer_x_axis(ax)
            ax.set_xlabel("Samples, $n$")
            ax.set_ylabel("Normalized frequency, $f / f_s$")
        ax.set_title("Spectrogram")



import numpy as np

import sdr


def test_encode_1():
    """
    MATLAB:
        >> diff_enc = comm.DifferentialEncoder;
        >> x = randi([0 1], 10, 1); x
        >> diff_enc(x)
    """
    x = [0, 1, 0, 0, 1, 1, 0, 0, 0, 0]
    y = sdr.diff_encode(x)
    y_truth = [0, 1, 1, 1, 0, 1, 1, 1, 1, 1]
    assert np.array_equal(y, y_truth)


# TODO: Add more MATLAB tests



import numpy as np

import sdr


def test_simple():
    map = np.array([0, 3, 1, 2])
    inverse_map = np.array([0, 2, 3, 1])
    interleaver = sdr.Interleaver(map)

    assert len(interleaver) == 4
    assert np.array_equal(interleaver.map, map)
    assert np.array_equal(interleaver.inverse_map, inverse_map)

    x_truth = np.arange(16)
    y_truth = np.array([0, 2, 3, 1, 4, 6, 7, 5, 8, 10, 11, 9, 12, 14, 15, 13])

    y = interleaver.interleave(x_truth)
    assert np.array_equal(y, y_truth)

    x = interleaver.deinterleave(y)
    assert np.array_equal(x, x_truth)



import galois
import numpy as np

import sdr


def verify_code(degree, poly, code_truth):
    code_truth = np.array(code_truth)

    decimal = sdr.m_sequence(degree, poly=poly, output="decimal")
    assert isinstance(decimal, np.ndarray)
    assert np.array_equal(decimal, code_truth)

    field = sdr.m_sequence(degree, poly=poly, output="field")
    assert isinstance(field, galois.FieldArray)
    assert np.array_equal(field, code_truth)

    bipolar = sdr.m_sequence(degree, poly=poly, output="bipolar")
    assert isinstance(bipolar, np.ndarray)
    assert np.array_equal(bipolar, 1 - 2 * code_truth)


def test_degree_2():
    """
    MATLAB:
        >> pn = comm.PNSequence('Polynomial', [2 1 0], 'InitialConditions', [0 1], 'SamplesPerFrame', 2^2 - 1);
        >> pn()
    """
    poly = galois.Poly.Degrees([2, 1, 0])
    code_truth = np.array([1, 0, 1])
    verify_code(2, poly, code_truth)


def test_degree_3():
    """
    MATLAB:
        >> pn = comm.PNSequence('Polynomial', [3 1 0], 'InitialConditions', [0 0 1], 'SamplesPerFrame', 2^3 - 1);
        >> pn()
    """
    poly = galois.Poly.Degrees([3, 1, 0])
    code_truth = np.array([1, 0, 0, 1, 0, 1, 1])
    verify_code(3, poly, code_truth)


def test_degree_4():
    """
    MATLAB:
        >> pn = comm.PNSequence('Polynomial', [4 1 0], 'InitialConditions', [0 0 0 1], 'SamplesPerFrame', 2^4 - 1);
        >> pn()
    """
    poly = galois.Poly.Degrees([4, 1, 0])
    code_truth = np.array([1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 1])
    verify_code(4, poly, code_truth)


def test_degree_5():
    """
    MATLAB:
        >> pn = comm.PNSequence('Polynomial', [5 2 0], 'InitialConditions', [0 0 0 0 1], 'SamplesPerFrame', 2^5 - 1);
        >> pn()
    """
    poly = galois.Poly.Degrees([5, 2, 0])
    code_truth = np.array([1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0])
    verify_code(5, poly, code_truth)


def test_degree_6():
    """
    MATLAB:
        >> pn = comm.PNSequence('Polynomial', [6 1 0], 'InitialConditions', [0 0 0 0 0 1], 'SamplesPerFrame', 2^6 - 1);
        >> pn()
    """
    poly = galois.Poly.Degrees([6, 1, 0])
    code_truth = np.array([1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1])  # fmt: skip
    verify_code(6, poly, code_truth)



"""
A module containing measurement functions related to modulation.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.integrate
from typing_extensions import Literal

from .._helper import export, verify_arraylike, verify_literal, verify_scalar
from ._power import average_power


@export
def evm(
    x_hat: npt.ArrayLike,
    ref: npt.ArrayLike,
    norm: Literal["average-power-ref", "average-power", "peak-power"] = "average-power-ref",
    output: Literal["rms", "all"] | float = "rms",
) -> float:
    r"""
    Measures the error-vector magnitude (EVM) of the complex symbols $\hat{x}[k]$.

    Arguments:
        x_hat: The complex symbols $\hat{x}[k]$ to be measured.
        ref: The complex reference symbols $x[k]$. This can be the noiseless transmitted symbols or the
            modulation's symbol map.
        norm: The normalization source used in the EVM calculation.

            - `"average-power-ref"`: The average power of the reference symbols $x[k]$.

            $$P_{\text{ref}} = \frac{1}{N} \sum_{k=0}^{N-1} \left| x[k] \right|^2$$

            - `"average-power"`: The average power of the received symbols $\hat{x}[k]$.

            $$P_{\text{ref}} = \frac{1}{N} \sum_{k=0}^{N-1} \left| \hat{x}[k] \right|^2$$

            - `"peak-power"`: The peak power of the received symbols $\hat{x}[k]$.

            $$P_{\text{ref}} = \text{max} \left| \hat{x}[k] \right|^2$$

        output: The output type of the EVM calculation.

            - `"rms"`: The root-mean-square (RMS) EVM.

            $$
            \text{EVM}_{\text{RMS}} =
            100 \sqrt{\frac{\frac{1}{N} \sum_{k=0}^{N-1} \left| \hat{x}[k] - x[k] \right|^2}{P_{\text{ref}}}}
            $$

            - `"all"`: The instantaneous EVM for each symbol.

            $$
            \text{EVM}_{k} =
            100 \sqrt{\frac{\left| \hat{x}[k] - x[k] \right|^2}{P_{\text{ref}}}}
            $$

            - `float`: The RMS EVM for the given percentile (0 - 100).

    Examples:
        Create QPSK symbols with $E_s/N_0$ of 20 dB.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, psk.order, 1000); \
            x = psk.map_symbols(s); \
            x_hat = sdr.awgn(x, 20)

            @savefig sdr_evm_1.png
            plt.figure(); \
            sdr.plot.constellation(x_hat, label=r"$\hat{x}[k]$"); \
            sdr.plot.symbol_map(psk.symbol_map, label=r"Reference"); \
            plt.title("QPSK Constellation at 20 dB $E_s/N_0$");

        Measure the RMS EVM, normalizing with the average power of the reference symbols.
        Either the symbol map or noiseless transmitted symbols may be passed.

        .. ipython:: python

            sdr.evm(x_hat, psk.symbol_map)
            sdr.evm(x_hat, x)

        Measure the RMS EVM, normalizing with the average power of the received symbols.

        .. ipython:: python

            sdr.evm(x_hat, psk.symbol_map, norm="average-power")

        Measure the RMS EVM, normalizing with the peak power of the received symbols.

        .. ipython:: python

            sdr.evm(x_hat, psk.symbol_map, norm="peak-power")

        Measure the 95th percentile EVM.

        .. ipython:: python

            sdr.evm(x_hat, psk.symbol_map, output=95)

        Measure the instantaneous EVM for each symbol.

        .. ipython:: python

            inst_evm = sdr.evm(x_hat, psk.symbol_map, output="all")

            @savefig sdr_evm_2.png
            plt.figure(); \
            plt.hist(inst_evm, bins=20); \
            plt.xlabel("RMS EVM (%)"); \
            plt.ylabel("Count"); \
            plt.title("EVM Histogram");

    Group:
        measurement-modulation
    """
    x_hat = verify_arraylike(x_hat, complex=True)
    ref = verify_arraylike(ref, complex=True)
    norm = verify_literal(norm, ["average-power-ref", "average-power", "peak-power"])

    if norm == "average-power-ref":
        ref_power = average_power(ref)
    elif norm == "average-power":
        ref_power = average_power(x_hat)
    elif norm == "peak-power":
        ref_power = np.max(np.abs(x_hat) ** 2)

    if ref.shape == x_hat.shape:
        # Compute the error vectors to each reference symbol
        error_vectors = x_hat - ref
    else:
        # The reference symbols are the symbol map. We must first determine the most likely reference symbol for each
        # received symbol. Then we must compute the error vectors to those symbols.
        symbol_map = ref
        all_error_vectors = np.subtract.outer(x_hat, symbol_map)
        s_hat = np.argmin(np.abs(all_error_vectors), axis=-1)
        error_vectors = x_hat - symbol_map[s_hat]

    if output == "rms":
        rms_evm = 100 * np.sqrt(average_power(error_vectors) / ref_power)
        return rms_evm

    inst_evm = 100 * np.sqrt(np.abs(error_vectors) ** 2 / ref_power)
    if output == "all":
        return inst_evm

    verify_scalar(output, float=True, inclusive_min=0, inclusive_max=100)
    perc_evm = np.percentile(inst_evm, output)
    return perc_evm


@export
def rms_bandwidth(
    x: npt.ArrayLike,
    sample_rate: float = 1.0,
) -> float:
    r"""
    Measures the RMS bandwidth $B_{\text{rms}}$ of the signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to measure.

            .. note::
                For best measurement performance, the time-domain signal should be very long. This allows for
                more averaging of its power spectra.

        sample_rate: The sample rate $f_s$ in samples/s.

    Returns:
        The RMS signal bandwidth $B_{\text{rms}}$ in Hz.

    See Also:
        sdr.toa_crlb, sdr.tdoa_crlb

    Notes:
        The root-mean-square (RMS) bandwidth $B_{\text{rms}}$ is calculated by

        $$
        B_{\text{rms}} = \sqrt{\frac
        {\int_{-\infty}^{\infty} (f - \mu_f)^2 \cdot S(f - \mu_f) \, df}
        {\int_{-\infty}^{\infty} S(f - \mu_f) \, df}
        }
        $$

        where $S(f)$ is the power spectral density (PSD) of the signal $x[n]$. The RMS bandwidth is measured about the
        centroid of the spectrum

        $$
        \mu_f = \frac
        {\int_{-\infty}^{\infty} f \cdot S(f) \, df}
        {\int_{-\infty}^{\infty} S(f) \, df} .
        $$

        The RMS bandwidth is a measure of the energy spread of the spectrum about the centroid. For a rectangular
        spectrum, the RMS bandwidth is $B_{\text{rms}} = B_s / \sqrt{12}$.

    Examples:
        Calculate the RMS bandwidth of a signal with a rectangular spectrum with normalized bandwidth of 1.

        .. ipython:: python

            symbol_rate = 1  # symbols/s
            symbol_rate / np.sqrt(12)

        Create a BPSK signal with a rectangular pulse shape. Note, the time-domain pulse shape is rectangular, but the
        spectrum is sinc-shaped. Measure the RMS bandwidth of the signal and compare it to the ideal rectangular
        spectrum.

        .. ipython:: python

            psk = sdr.PSK(2, pulse_shape="rect")
            symbols = np.random.randint(0, psk.order, 10_000)
            x_rect = psk.modulate(symbols)
            sdr.rms_bandwidth(x_rect, sample_rate=symbol_rate * psk.sps)

        Make the same measurements with square-root raised cosine (SRRC) pulse shaping. The SRRC spectrum is narrower
        and, therefore, closer to the rectangular spectrum.

        .. ipython:: python

            psk = sdr.PSK(2, pulse_shape="srrc")
            symbols = np.random.randint(0, psk.order, 10_000)
            x_srrc = psk.modulate(symbols)
            sdr.rms_bandwidth(x_srrc, sample_rate=symbol_rate * psk.sps)

        Plot the power spectral density (PSD) of the rectangular and SRRC pulse-shaped signals.

        .. ipython:: python

            @savefig sdr_rms_bandwidth_1.png
            plt.figure(); \
            sdr.plot.periodogram(x_rect, sample_rate=symbol_rate * psk.sps, label="Rectangular"); \
            sdr.plot.periodogram(x_srrc, sample_rate=symbol_rate * psk.sps, label="SRRC");

    Group:
        measurement-modulation
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    verify_scalar(sample_rate, float=True, positive=True)

    f, psd = scipy.signal.welch(
        x, fs=sample_rate, detrend=False, return_onesided=False, scaling="density", average="mean"
    )

    # Shift so that the frequencies increase monotonically
    f = np.fft.fftshift(f)
    psd = np.fft.fftshift(psd)

    # Calculate the centroid of the PSD
    f_mean = scipy.integrate.simpson(f * psd, x=f)
    f_mean /= scipy.integrate.simpson(psd, x=f)
    f -= f_mean

    # Calculate the RMS bandwidth
    ms_bandwidth = scipy.integrate.simpson(f**2 * psd, x=f)
    ms_bandwidth /= scipy.integrate.simpson(psd, x=f)
    rms_bandwidth = np.sqrt(float(ms_bandwidth))

    return rms_bandwidth


@export
def rms_integration_time(
    x: npt.ArrayLike,
    sample_rate: float = 1.0,
) -> float:
    r"""
    Measures the RMS integration time $T_{\text{rms}}$ of the signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$ to measure.

            .. note::
                For best measurement performance, the time-domain signal should be very oversampled. This allows
                for better time precision.

        sample_rate: The sample rate $f_s$ in samples/s.

    Returns:
        The RMS integration time $T_{\text{rms}}$ in seconds.

    See Also:
        sdr.foa_crlb, sdr.fdoa_crlb

    Notes:
        The root-mean-square (RMS) integration time $T_{\text{rms}}$ is calculated by

        $$
        T_{\text{rms}} = \sqrt{\frac
        {\int_{-\infty}^{\infty} (t - \mu_t)^2 \cdot \left| x(t - \mu_t) \right|^2 \, dt}
        {\int_{-\infty}^{\infty} \left| x(t - \mu_t) \right|^2 \, dt}
        }
        $$

        where $x(t)$ is the continuous-time signal that was discretely sampled as $x[n]$. The RMS integration time is
        measured about the centroid of the signal

        $$
        \mu_t = \frac
        {\int_{-\infty}^{\infty} t \cdot \left| x(t) \right|^2 \, dt}
        {\int_{-\infty}^{\infty} \left| x(t) \right|^2 \, dt} .
        $$

        The RMS integration time is a measure of the energy spread of the signal about the centroid. For a rectangular
        signal, the RMS integration time is $T_{\text{rms}} = T / \sqrt{12}$.

    Examples:
        Calculate the RMS integration time of a signal with a rectangular envelope and duration of 1 second.

        .. ipython:: python

            symbol_rate = 100  # symbols/s
            sps = 100  # samples/symbol
            sample_rate = symbol_rate * sps  # samples/s
            n_symbols = symbol_rate  # Make a 1-second long signal
            t_s = n_symbols / symbol_rate  # Integration time (s)
            t_s / np.sqrt(12)

        Create a BPSK signal with a rectangular pulse shape. Measure the RMS integration time of the signal and compare
        it to the ideal rectangular envelope.

        .. ipython:: python

            psk = sdr.PSK(2, pulse_shape="rect", sps=sps)
            symbols = np.random.randint(0, psk.order, n_symbols)
            x_rect = psk.modulate(symbols).real
            sdr.rms_integration_time(x_rect, sample_rate=sample_rate)

            @savefig sdr_rms_integration_time_1.png
            plt.figure(); \
            sdr.plot.time_domain(x_rect, sample_rate=sample_rate, label="Rectangular");

        Make the same measurements with square-root raised cosine (SRRC) pulse shaping.

        .. ipython:: python

            psk = sdr.PSK(2, pulse_shape="srrc", sps=sps)
            symbols = np.random.randint(0, psk.order, n_symbols)
            x_srrc = psk.modulate(symbols).real
            sdr.rms_integration_time(x_srrc, sample_rate=sample_rate)

            @savefig sdr_rms_integration_time_2.png
            plt.figure(); \
            sdr.plot.time_domain(x_srrc, sample_rate=sample_rate, label="SRRC");

        For a given transmit energy, the RMS integration time is improved by increasing the energy at the edges of the
        signal. This can be achieved by applying a parabolic envelope to the BPSK signal. The energy of the signals
        is normalized. Notice the RMS integration time increases by 50% for the same transmit energy and duration.

        .. ipython:: python

            x_srrc_env = x_srrc * np.linspace(-1, 1, len(x_srrc))**2
            x_srrc_env *= np.sqrt(sdr.energy(x_srrc) / sdr.energy(x_srrc_env))
            sdr.rms_integration_time(x_srrc_env, sample_rate=sample_rate)

            @savefig sdr_rms_integration_time_3.png
            plt.figure(); \
            sdr.plot.time_domain(x_srrc, sample_rate=sample_rate, label="SRRC"); \
            sdr.plot.time_domain(x_srrc_env, sample_rate=sample_rate, label="SRRC + Parabolic Envelope");

    Group:
        measurement-modulation
    """
    x = verify_arraylike(x, complex=False, ndim=1)
    verify_scalar(sample_rate, float=True, positive=True)

    t = np.arange(x.size) / sample_rate

    # Calculate the centroid of the signal
    t_mean = scipy.integrate.simpson(t * np.abs(x) ** 2, x=t)
    t_mean /= scipy.integrate.simpson(np.abs(x) ** 2, x=t)
    t -= t_mean

    # Calculate the RMS integration time
    ms_integration_time = scipy.integrate.simpson(t**2 * np.abs(x) ** 2, x=t)
    ms_integration_time /= scipy.integrate.simpson(np.abs(x) ** 2, x=t)
    rms_integration_time = np.sqrt(float(ms_integration_time))

    return rms_integration_time



"""
A module for closed-loop phase-locked loops (PLLs) analysis.
"""

from __future__ import annotations

import numpy as np

from .._conversion import linear
from .._filter import IIR
from .._helper import export, verify_scalar
from ._loop_filter import LoopFilter


@export
class ClosedLoopPLL:
    r"""
    A class that defines the performance of a closed-loop PLL.

    Note:
        This class is meant for performance analysis only.

    Notes:
        .. code-block:: text
            :caption: Closed-Loop PLL Block Diagram

                     bb[n]    phase_err[n]
                         +-----+    +----+
            x[n] -->X--->| PED |--->| LF |---+
                    ^    +-----+    +----+   |
                    |                        | phase_est[n]
                    |    +-----+             |
             lo[n]  +----| NCO |<------------+
                         +-----+

            x[n] = Input signal
            bb[n] = Baseband signal
            phase_err[n] = Phase error signal
            phase_est[n] = Phase estimate signal
            lo[n] = Local oscillator signal
            PED = Phase error detector
            LF = Loop filter
            NCO = Numerically controlled oscillator
            X = Complex multiplier

        The transfer function of the 2nd order, proportional-plus-integrator loop filter is

        $$H_{LF}(z) = K_1 + K_2 \frac{ 1 }{ 1 - z^{-1}} = \frac{(K_1 + K_2) - K_1 z^{-1}}{1 - z^{-1}} .$$

        The transfer function of the NCO is

        $$H_{NCO}(z) = K_0 \frac{z^{-1}}{1 - z^{-1}} .$$

        The closed-loop transfer function of the PLL is

        $$
        H_{PLL}(z) = \frac{K_p K_0 (K_1 + K_2) z^{-1} - K_p K_0 K_1 z^{-2}}
        {1 - 2 (1 - \frac{1}{2} K_p K_0 (K_1 + K_2) z^{-1} + (1 - K_p K_0 K_1) z^{-2} } .
        $$

    References:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Appendix C: Phase Locked Loops.

    Examples:
        See the :ref:`phase-locked-loop` example.

    Group:
        pll
    """

    def __init__(
        self,
        noise_bandwidth: float,
        damping_factor: float,
        K0: float = 1.0,
        Kp: float = 1.0,
        sample_rate: float = 1.0,
    ):
        r"""
        Creates a closed-loop PLL analysis object.

        Arguments:
            noise_bandwidth: The normalized noise bandwidth $B_n T$ of the loop filter,
                where $B_n$ is the noise bandwidth in Hz and $T$ is the sampling period in seconds.
            damping_factor: The damping factor $\zeta$ of the loop filter. $\zeta = 1$ is critically damped,
                $\zeta < 1$ is underdamped, and $\zeta > 1$ is overdamped.
            K0: The NCO gain $K_0$.
            Kp: The gain $K_p$ of the phase error detector (PED) or time error detector (TED).
            sample_rate: The sample rate $f_s$ of the PLL in Hz.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        self._BnT = verify_scalar(noise_bandwidth, float=True, positive=True)
        self._zeta = verify_scalar(damping_factor, float=True, positive=True)
        self._K0 = verify_scalar(K0, float=True, positive=True)
        self._Kp = verify_scalar(Kp, float=True, positive=True)
        self._sample_rate = verify_scalar(sample_rate, float=True, positive=True)

        lf = LoopFilter(noise_bandwidth, damping_factor, K0, Kp)
        K1 = lf.proportional_gain
        K2 = lf.integral_gain

        # b0 = 0
        b1 = Kp * K0 * (K1 + K2)
        b2 = -Kp * K0 * K1

        a0 = 1
        a1 = -2 * (1 - 0.5 * Kp * K0 * (K1 + K2))
        a2 = 1 - Kp * K0 * K1

        # Create an IIR filter that represents the closed-loop transfer function of the PLL
        # self._iir = IIR([b0, b1, b2], [a0, a1, a2])
        self._iir = IIR([b1, b2], [a0, a1, a2])  # Suppress warning about leading zero

        self._K1 = K1
        self._K2 = K2

    def phase_lock_time(self) -> float:
        r"""
        Returns the phase lock time of the PLL.

        $$
        T_{PL} = \frac{1.3}{B_n}
        $$

        Returns:
            The time $T_{PL}$ it takes the PLL to lock onto the input signal's phase in seconds.

        References:
            - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Equation C.40.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return 1.3 / self.Bn

    def frequency_lock_time(self, freq_offset: float) -> float:
        r"""
        Returns the frequency lock time of the PLL.

        $$
        T_{FL} = 4 \frac{(\Delta f)^2}{B_n^3}
        $$

        Arguments:
            freq_offset: The frequency offset $\Delta f$ of the input signal in Hz.

        Returns:
            The time $T_{FL}$ it takes the PLL to lock onto the input signal's frequency in seconds.

        References:
            - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Equation C.39.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return 4 * freq_offset**2 / self.Bn**3

    def lock_time(self, freq_offset: float) -> float:
        r"""
        Returns the lock time of the PLL.

        $$
        T_{LOCK} = T_{PL} + T_{FL} = \frac{1.3}{B_n} + 4 \frac{(\Delta f)^2}{B_n^3}
        $$

        Arguments:
            freq_offset: The frequency offset $\Delta f$ of the input signal in Hz.

        Returns:
            The time $T_{LOCK}$ it takes the PLL to lock onto the input signal's phase and frequency in seconds.

        References:
            - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Equation C.38.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self.phase_lock_time() * self.frequency_lock_time(freq_offset)

    def phase_error_variance(self, cn0: float) -> float:
        r"""
        Returns the variance of the phase error of the PLL in steady state.

        $$
        \sigma_{\theta_e}^2 = \frac{N_0 B_n}{C}
        $$

        Arguments:
            cn0: The carrier-to-noise density ratio $C/N_0$ of the input signal in dB-Hz.

        Returns:
            The variance of the phase error $\sigma_{\theta_e}^2$ of the PLL in radians^2.

        References:
            - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Equation C.43.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        cn0_linear = linear(cn0)
        return self.Bn / cn0_linear

    @property
    def sample_rate(self) -> float:
        """
        The sample rate $f_s$ of the PLL in samples/s.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._sample_rate

    @property
    def BnT(self) -> float:
        """
        The normalized noise bandwidth $B_n T$ of the PLL.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._BnT

    @property
    def Bn(self) -> float:
        """
        The noise bandwidth $B_n$ of the PLL in Hz.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self.BnT * self.sample_rate

    @property
    def zeta(self) -> float:
        r"""
        The damping factor $\zeta$ of the PLL.

        A damping factor of 1 is critically damped, less than 1 is underdamped, and greater than 1 is overdamped.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._zeta

    @property
    def K0(self) -> float:
        """
        The NCO gain $K_0$.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._K0

    @property
    def Kp(self) -> float:
        """
        The phase error detector (PED) gain $K_p$.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._Kp

    @property
    def K1(self) -> float:
        """
        The proportional gain $K_1$ of the loop filter.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._K1

    @property
    def K2(self) -> float:
        """
        The integral gain $K_2$ of the loop filter.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._K2

    @property
    def iir(self) -> IIR:
        """
        The IIR filter that represents the closed-loop transfer function of the PLL.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._iir

    @property
    def omega_n(self) -> float:
        r"""
        The natural frequency $\omega_n$ of the PLL in radians/s.

        References:
            - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Equation C.33.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return np.sqrt(self.K0 * self.Kp * self.K2)

    @property
    def omega_3dB(self) -> float:
        r"""
        The 3-dB bandwidth $\omega_{3\textrm{dB}}$ of the PLL in radians/s.

        References:
            - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Equation C.34.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self.omega_n * np.sqrt(1 + 2 * self.zeta**2 + np.sqrt((1 + 2 * self.zeta**2) ** 2 + 1))



"""
A module for discrete-time loop filters.
"""

from __future__ import annotations

import numpy.typing as npt

from .._filter import IIR
from .._helper import export, verify_scalar


@export
class LoopFilter:
    r"""
    Implements a 2nd order, proportional-plus-integrator (PPI) loop filter.

    Notes:
        .. code-block:: text
            :caption: Proportional-Plus-Integral Loop Filter Block Diagram

                       +----+
                   +-->| K1 |---------------------+
                   |   +----+                     |
            x[n] --+                              @--> y[n]
                   |   +----+                     |
                   +-->| K2 |-->@--------------+--+
                       +----+   ^              |
                                |   +------+   |
                                +---| z^-1 |<--+
                                    +------+

            x[n] = Input signal
            y[n] = Output signal
            K1 = Proportional gain
            K2 = Integral gain
            z^-1 = Unit delay
            @ = Adder

        The transfer function of the loop filter is

        $$H(z) = K_1 + K_2 \frac{ 1 }{ 1 - z^{-1}} = \frac{(K_1 + K_2) - K_1 z^{-1}}{1 - z^{-1}} .$$

        The second-order proportional-plus-integrator loop filter can track a constant phase error
        and/or frequency error to zero. It cannot, however, track a constant chirp (frequency ramp)
        to zero.

    References:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Appendix C: Phase Locked Loops.

    Examples:
        See the :ref:`phase-locked-loop` example.

    Group:
        pll
    """

    def __init__(
        self,
        noise_bandwidth: float,
        damping_factor: float,
        K0: float = 1.0,
        Kp: float = 1.0,
    ):
        r"""
        Creates a 2nd order, proportional-plus-integrator (PPI) loop filter.

        Arguments:
            noise_bandwidth: The normalized noise bandwidth $B_n T$ of the loop filter,
                where $B_n$ is the noise bandwidth in Hz and $T$ is the sampling period in seconds.
            damping_factor: The damping factor $\zeta$ of the loop filter. $\zeta = 1$ is critically damped,
                $\zeta < 1$ is underdamped, and $\zeta > 1$ is overdamped.
            K0: The NCO gain $K_0$.
            Kp: The gain $K_p$ of the phase error detector (PED) or time error detector (TED).

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        BnT = verify_scalar(noise_bandwidth, float=True, positive=True)
        zeta = verify_scalar(damping_factor, float=True, positive=True)
        verify_scalar(K0, float=True, positive=True)
        verify_scalar(Kp, float=True, positive=True)

        # Equation C.57, page 736
        theta_n = BnT / (zeta + 1 / (4 * zeta))

        # Equation C.58, page 737
        K1 = 4 * zeta * theta_n / (1 + 2 * zeta * theta_n + theta_n**2) / K0 / Kp
        K2 = 4 * theta_n**2 / (1 + 2 * zeta * theta_n + theta_n**2) / K0 / Kp

        b = [K1 + K2, -K1]
        a = [1, -1]

        self._BnT = BnT
        self._zeta = zeta
        self._K1 = K1
        self._K2 = K2
        self._iir = IIR(b, a, streaming=True)

        self.reset()

    def reset(self) -> None:
        """
        Resets the loop filter.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        self.iir.reset()

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray:
        """
        Filters the input signal $x[n]$.

        Arguments:
            x: The input signal $x[n]$.

        Returns:
            The filtered output signal $y[n]$.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self.iir(x)

    @property
    def noise_bandwidth(self) -> float:
        """
        The normalized noise bandwidth $B_n T$ of the loop filter.

        $B_n$ is the noise bandwidth in Hz and $T$ is the sampling period in seconds.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._BnT

    @property
    def damping_factor(self) -> float:
        r"""
        The damping factor $\zeta$ of the loop filter.

        $\zeta = 1$ is critically damped, $\zeta < 1$ is underdamped, and $\zeta > 1$ is overdamped.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._zeta

    @property
    def proportional_gain(self) -> float:
        """
        The proportional gain $K_1$ of the loop filter.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._K1

    @property
    def integral_gain(self) -> float:
        """
        The integral gain $K_2$ of the loop filter.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._K2

    @property
    def iir(self) -> IIR:
        """
        The underlying IIR filter used to implement the loop filter.

        Examples:
            See the :ref:`phase-locked-loop` example.
        """
        return self._iir
